import logging
import shutil
from itertools import chain
from pathlib import Path

import numpy as np
import pandas as pd

from immuneML.data_model.SequenceParams import RegionType, Chain
from immuneML.data_model.bnp_util import write_yaml, read_yaml, get_sequence_field_name
from immuneML.data_model.datasets.ElementDataset import SequenceDataset
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.environment.SequenceType import SequenceType
from immuneML.ml_methods.generative_models.GenerativeModel import GenerativeModel
from immuneML.ml_methods.pytorch_implementations.SimpleLSTMGenerator import SimpleLSTMGenerator
from immuneML.ml_methods.util.pytorch_util import store_weights
from immuneML.util.Logger import print_log
from immuneML.util.PathBuilder import PathBuilder


class SimpleLSTM(GenerativeModel):
    """
    This is a simple generative model for receptor sequences based on LSTM.

    Similar models have been proposed in:

    Akbar, R. et al. (2022). In silico proof of principle of machine learning-based antibody design at unconstrained scale. mAbs, 14(1), 2031482. https://doi.org/10.1080/19420862.2022.2031482

    Saka, K. et al. (2021). Antibody design using LSTM based deep generative model from phage display library for affinity maturation. Scientific Reports, 11(1), Article 1. https://doi.org/10.1038/s41598-021-85274-7


    **Specification arguments:**

    - sequence_type (str): whether the model should work on amino_acid or nucleotide level

    - hidden_size (int): how many LSTM cells should exist per layer

    - num_layers (int): how many hidden LSTM layers should there be

    - num_epochs (int): for how many epochs to train the model

    - learning_rate (float): what learning rate to use for optimization

    - batch_size (int): how many examples (sequences) to use for training for one batch

    - embed_size (int): the dimension of the sequence embedding

    - temperature (float): a higher temperature leads to faster yet more unstable learning

    - prime_str (str): the initial sequence to start generating from

    - seed (int): random seed for the model or None

    - iter_to_report (int): number of epochs between training progress reports


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            ml_methods:
                my_simple_lstm:
                    sequence_type: amino_acid
                    hidden_size: 50
                    num_layers: 1
                    num_epochs: 5000
                    learning_rate: 0.001
                    batch_size: 100
                    embed_size: 100


    """

    @classmethod
    def load_model(cls, path: Path):
        assert path.exists(), f"{cls.__name__}: {path} does not exist."

        model_overview_file = path / 'model_overview.yaml'
        state_dict_file = path / 'state_dict.yaml'

        for file in [model_overview_file, state_dict_file]:
            assert file.exists(), f"{cls.__name__}: {file} is not a file."

        model_overview = read_yaml(model_overview_file)
        lstm = SimpleLSTM(**{k: v for k, v in model_overview.items() if k != 'type'})
        lstm._model = lstm.make_new_model(state_dict_file)
        return lstm

    def __init__(self, locus: str, sequence_type: str, hidden_size: int, learning_rate: float, num_epochs: int,
                 batch_size: int, num_layers: int, embed_size: int, temperature, device: str, name=None,
                 region_type: str = RegionType.IMGT_CDR3.name, prime_str: str = "C", window_size: int = 64,
                 seed: int = None, iter_to_report: int = 1):

        super().__init__(locus, region_type=RegionType.get_object(region_type), name=name, seed=seed)
        self._model = None
        self.sequence_type = SequenceType[sequence_type.upper()] if sequence_type else SequenceType.AMINO_ACID
        self.hidden_size = hidden_size
        self.learning_rate = learning_rate
        self.num_layers = num_layers
        self.num_epochs = num_epochs
        self.batch_size = batch_size
        self.embed_size = embed_size
        self.temperature = temperature
        self.prime_str = prime_str
        self.window_size = window_size
        self.device = device
        self.iter_to_report = iter_to_report
        self.unique_letters = EnvironmentSettings.get_sequence_alphabet(self.sequence_type) + ["*"]
        self.num_letters = len(self.unique_letters)
        self.letter_to_index = {letter: i for i, letter in enumerate(self.unique_letters)}
        self.index_to_letter = {i: letter for letter, i in self.letter_to_index.items()}
        self.loss_summary_path = None

    def make_new_model(self, state_dict_file: Path = None):
        from torch import as_tensor
        model = SimpleLSTMGenerator(input_size=self.num_letters, hidden_size=self.hidden_size,
                                    embed_size=self.embed_size, output_size=self.num_letters,
                                    batch_size=self.batch_size, device=self.device)

        if isinstance(state_dict_file, Path) and state_dict_file.is_file():
            state_dict = read_yaml(state_dict_file)
            state_dict = {k: as_tensor(v, device=self.device) for k, v in state_dict.items()}
            model.load_state_dict(state_dict)

        model.to(self.device)

        return model

    def _log_training_progress(self, loss_summary, epoch, loss):
        if (epoch + 1) % self.iter_to_report == 0:
            message = f"{SimpleLSTM.__name__}: Epoch [{epoch + 1}/{self.num_epochs}]: loss: {loss:.4f}"
            print_log(message, True)

            loss_summary['loss'].append(loss)
            loss_summary['epoch'].append(epoch + 1)
        return loss_summary

    def fit(self, data, path: Path = None):
        import torch
        from torch import nn, optim

        self.set_locus(data)

        if self.seed is not None:
            torch.manual_seed(self.seed)

        data_loader = self._encode_dataset(data)
        model = self.make_new_model()
        model.train()

        criterion = nn.CrossEntropyLoss(reduction='mean')
        optimizer = optim.Adam(model.parameters(), self.learning_rate)
        loss_summary = {"loss": [], "epoch": []}

        for epoch in range(self.num_epochs):
            epoch_loss = 0.
            num_batches = 0

            for x_batch, y_batch in data_loader:
                x_batch, y_batch = x_batch.to(self.device), y_batch.to(self.device)

                # Initialize state for each batch
                state = model.init_zero_state(x_batch.size(0))

                optimizer.zero_grad()

                outputs, _ = model(x_batch, state)
                outputs = outputs.reshape(-1, outputs.size(-1))
                y_batch = y_batch.reshape(-1)

                loss = criterion(outputs, y_batch)

                loss.backward()
                torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=5)
                optimizer.step()

                epoch_loss += loss.item()
                num_batches += 1

            avg_epoch_loss = epoch_loss / num_batches
            loss_summary = self._log_training_progress(loss_summary, epoch, avg_epoch_loss)

        self._log_training_summary(loss_summary, path)
        self._model = model

    def _log_training_summary(self, loss_summary, path):
        if path is not None:
            try:
                PathBuilder.build(path)
                self.loss_summary_path = path / 'loss_summary.csv'
                pd.DataFrame(loss_summary).to_csv(str(self.loss_summary_path), index=False)
            except Exception as e:
                logging.error(f"{SimpleLSTM.__name__}: failed to save loss summary: {e};\n{loss_summary}")

    def _encode_dataset(self, dataset: SequenceDataset):
        from torch import as_tensor
        from torch.utils.data import DataLoader, TensorDataset

        seq_col = get_sequence_field_name(self.region_type, self.sequence_type)
        sequences = dataset.get_attribute(seq_col).tolist()

        # Flatten all sequences into one long sequence with end tokens
        sequences = list(chain.from_iterable(
            [[self.letter_to_index[letter] for letter in seq] + [self.letter_to_index['*']] for seq in sequences]))

        sequences = as_tensor(sequences, device=self.device).long()

        # Create overlapping windows of size window_size
        # stride=1 means each window overlaps with previous window by window_size-1 elements
        windows = sequences.unfold(0, self.window_size, 1)

        # Create input-target pairs from windows
        x = windows[:, :-1]  # All but last character of each window
        y = windows[:, 1:]  # All but first character of each window

        return DataLoader(TensorDataset(x, y), shuffle=True, batch_size=self.batch_size)

    def is_same(self, model) -> bool:
        raise NotImplementedError

    def generate_sequences(self, count: int, seed: int, path: Path, sequence_type: SequenceType, compute_p_gen: bool,
                           max_failed_trials: int = 1000):
        import torch

        torch.manual_seed(seed)

        self._model.eval()
        input_vector = torch.as_tensor([self.letter_to_index[letter] for letter in self.prime_str],
                                       device=self.device).long()
        predicted = self.prime_str
        failed_trials = 0
        total_trials = 0

        with torch.no_grad():
            state = self._model.init_zero_state(batch_size=1)

            # Add sequence and batch dimensions for LSTM input
            for p in range(len(self.prime_str) - 1):
                inp = input_vector[p].unsqueeze(0).unsqueeze(0)
                _, state = self._model(inp, state)

            inp = input_vector[-1].unsqueeze(0).unsqueeze(0)
            gen_seq_count = 0

            while gen_seq_count < count and failed_trials < max_failed_trials:
                output, state = self._model(inp, state)

                scaled_logits = output[0, 0] / self.temperature
                scaled_logits = torch.clamp(scaled_logits, min=-100, max=100)
                output_dist = torch.nn.functional.softmax(scaled_logits, dim=0)

                try:
                    output_dist = output_dist + 1e-10
                    output_dist = output_dist / output_dist.sum()
                    top_i = torch.multinomial(output_dist, 1)[0].item()
                except RuntimeError as e:
                    logging.warning(f"{SimpleLSTM.__name__}: Error sampling from distribution: {e}; "
                                    f"Using argmax instead.")
                    logging.debug(f"{SimpleLSTM.__name__}: Distribution: {output_dist}\n"
                                  f"Sum: {output_dist.sum()}, Min: {output_dist.min()}, Max: {output_dist.max()}")
                    top_i = scaled_logits.argmax().item()

                predicted_char = self.index_to_letter[top_i]
                predicted += predicted_char
                inp = torch.as_tensor(self.letter_to_index[predicted_char], device=self.device).long()
                inp = inp.unsqueeze(0).unsqueeze(0)

                if predicted_char == "*":
                    last_seq = predicted.split('*')[-2]
                    if len(last_seq) > 0:
                        gen_seq_count += 1
                        print_log(f"Generated valid sequence {gen_seq_count}/{count}", True)

                    else:
                        failed_trials += 1
                        if failed_trials % 10 == 0:
                            print_log(f"Warning: LSTM model generated {failed_trials} empty sequences", True)

                total_trials += 1

        print_log(
            f"{SimpleLSTM.__name__} {self.name}: generated {gen_seq_count} sequences with {failed_trials} failed attempts.",
            True)

        dataset = self._export_dataset(predicted, path)
        return dataset

    def _export_dataset(self, predicted, path):
        sequences = [seq for seq in predicted.split('*') if len(seq) > 0]
        count = len(sequences)
        df = pd.DataFrame({get_sequence_field_name(self.region_type, self.sequence_type): sequences,
                           'locus': [self.locus.to_string() for _ in range(count)],
                           'gen_model_name': [self.name for _ in range(count)]})

        return SequenceDataset.build_from_partial_df(df, PathBuilder.build(path), 'synthetic_lstm_dataset',
                                                     {'gen_model_name': [self.name]}, {'gen_model_name': str})


    def compute_p_gens(self, sequences, sequence_type: SequenceType) -> np.ndarray:
        raise RuntimeError

    def compute_p_gen(self, sequence: dict, sequence_type: SequenceType) -> float:
        raise RuntimeError

    def can_compute_p_gens(self) -> bool:
        return False

    def can_generate_from_skewed_gene_models(self) -> bool:
        return False

    def generate_from_skewed_gene_models(self, v_genes: list, j_genes: list, seed: int, path: Path,
                                         sequence_type: SequenceType, batch_size: int, compute_p_gen: bool):
        raise RuntimeError

    def save_model(self, path: Path) -> Path:
        model_path = PathBuilder.build(path / 'model')

        skip_keys_for_export = ['_model', 'loss_summary_path', 'index_to_letter', 'letter_to_index',
                                'unique_letters', 'num_letters']
        write_yaml(filename=model_path / 'model_overview.yaml',
                   yaml_dict={**{k: v for k, v in vars(self).items() if k not in skip_keys_for_export},
                              **{'type': self.__class__.__name__, 'region_type': self.region_type.name,
                                 'sequence_type': self.sequence_type.name, 'locus': self.locus.name}})

        store_weights(self._model, model_path / 'state_dict.yaml')

        return Path(shutil.make_archive(str(path / f'trained_model_{self.name}'), 'zip', str(model_path))).absolute()



import logging
from dataclasses import dataclass, field, fields
from pathlib import Path
from typing import Any, Dict, List
from uuid import uuid4

from bionumpy import DNAEncoding, get_bufferclass_for_datatype, AminoAcidEncoding

from immuneML.data_model.AIRRSequenceSet import AIRRSequenceSet
from immuneML.data_model.SequenceParams import ChainPair, Chain, RegionType
from immuneML.data_model.bnp_util import get_field_type_from_values, \
    extend_dataclass_with_dynamic_fields, bnp_write_to_file, write_yaml, bnp_read_from_file, read_yaml, \
    build_dynamic_bnp_dataclass_obj
from immuneML.environment.SequenceType import SequenceType
from immuneML.util.PathBuilder import PathBuilder


@dataclass
class ReceptorSequence:
    sequence_id: str = ''
    sequence: DNAEncoding = ''
    sequence_aa: AminoAcidEncoding = ''
    productive: str = 'T'
    vj_in_frame: str = 'T'
    stop_codon: str = 'F'
    locus: str = ''
    locus_species: str = ''
    v_call: str = ''
    d_call: str = ''
    j_call: str = ''
    c_call: str = ''
    metadata: dict = field(default_factory=dict)
    duplicate_count: int = -1
    cell_id: str = ''

    @property
    def v_gene(self):
        if self.v_call and len(self.v_call) > 0:
            return self.v_call.split("*")[0]
        else:
            return ""

    def get_sequence(self, sequence_type: SequenceType = SequenceType.AMINO_ACID):
        return self.sequence_aa if sequence_type == SequenceType.AMINO_ACID else self.sequence

    def get_attribute(self, attr_name):
        try:
            if hasattr(self, attr_name):
                return getattr(self, attr_name)
            else:
                return self.metadata.get(attr_name)
        except KeyError as e:
            logging.error(f"ReceptorSequence object has no attribute {e}. In metadata, "
                          f"it has: {list(self.metadata.keys()) if isinstance(self.metadata, dict) else []}")
            raise

    def __post_init__(self):
        if self.sequence_id is None or self.sequence_id == "":
            self.sequence_id = uuid4().hex


@dataclass
class Receptor:
    chain_pair: ChainPair
    chain_1: ReceptorSequence
    chain_2: ReceptorSequence
    receptor_id: str = None
    cell_id: str = None
    metadata: dict = field(default_factory=dict)

    def __post_init__(self):
        new_metadata = {}
        for key, val in self.metadata.items():
            if isinstance(val, list) and len(val) == 1:
                new_metadata[key] = val[0]
            else:
                new_metadata[key] = val
        self.metadata = new_metadata

        if self.receptor_id is None:
            self.receptor_id = uuid4().hex

        if isinstance(self.chain_pair, str):
            self.chain_pair = ChainPair[self.chain_pair]

        setattr(self, Chain.get_chain(self.chain_pair.value[0]).name.lower(), self.chain_1)
        setattr(self, Chain.get_chain(self.chain_pair.value[1]).name.lower(), self.chain_2)


@dataclass
class Repertoire:
    data_filename: Path = None
    metadata_filename: Path = None
    metadata: dict = None
    identifier: str = None
    dynamic_fields: dict = None
    element_count: int = None

    def __post_init__(self):
        if not self.identifier:
            self.identifier = uuid4().hex
        if not self.metadata:
            self.metadata = read_yaml(self.metadata_filename)
        if not self.dynamic_fields and 'type_dict_dynamic_fields' in self.metadata:
            self.dynamic_fields = self.metadata.get('type_dict_dynamic_fields', {})
            for name, field_type in self.dynamic_fields.items():
                self.dynamic_fields[name] = AIRRSequenceSet.STR_TO_TYPE.get(field_type, field_type)

    @classmethod
    def build_from_dc_object(cls, path: Path, metadata: dict, filename_base: str = None, identifier: str = None,
                             data=None, type_dict: dict = None):
        identifier = uuid4().hex if identifier is None else identifier
        filename_base = filename_base if filename_base is not None else identifier
        data_filename = path / f"{filename_base}.tsv"

        bnp_write_to_file(data_filename, data)

        metadata_filename = path / f"{filename_base}.yaml"
        metadata = {} if metadata is None else metadata

        if not type_dict:
            type_dict = {f.name: f.type for f in fields(data)
                         if f.name not in [airr_field.name for airr_field in fields(AIRRSequenceSet)]}

        metadata['type_dict_dynamic_fields'] = {key: AIRRSequenceSet.TYPE_TO_STR[val] for key, val in type_dict.items()}
        write_yaml(metadata_filename, metadata)

        repertoire = Repertoire(data_filename, metadata_filename, metadata, identifier, type_dict, len(data))
        return repertoire

    @classmethod
    def build(cls, path: Path, metadata: dict, filename_base: str = None, identifier: str = None, **kwargs):

        bnp_dc, type_dict = build_dynamic_airr_sequence_set_dataclass(kwargs)
        el_count = list(kwargs.values())[0]
        el_count = len(el_count) if isinstance(el_count, list) else el_count.shape[0]
        kwargs_with_missing_fields = {
            **kwargs,
            **{k: [AIRRSequenceSet.get_neutral_value(t) for _ in range(el_count)]
               for k, t in AIRRSequenceSet.get_field_type_dict().items() if k not in kwargs}
        }
        data = bnp_dc(**kwargs_with_missing_fields)

        return Repertoire.build_from_dc_object(path, metadata, filename_base, identifier, data, type_dict)

    @classmethod
    def build_like(cls, repertoire: 'Repertoire', indices_to_keep, result_path: Path, filename_base: str):
        identifier = uuid4().hex
        filename_base = filename_base if filename_base is not None else identifier
        data = bnp_read_from_file(repertoire.data_filename, repertoire.buffer_type)
        data = data[indices_to_keep]

        PathBuilder.build(result_path)

        data_filename = result_path / f"{filename_base}.tsv"
        bnp_write_to_file(data_filename, data)

        metadata_filename = result_path / f"{filename_base}.yaml"
        metadata = read_yaml(repertoire.metadata_filename)
        write_yaml(metadata_filename, metadata)

        repertoire = Repertoire(data_filename, metadata_filename, metadata, identifier,
                                dynamic_fields=repertoire.dynamic_fields)
        repertoire.element_count = len(data)
        return repertoire

    @classmethod
    def build_from_sequences(cls, sequences: List[ReceptorSequence], result_path: Path, filename_base: str = None,
                             metadata: dict = None, region_type: RegionType = RegionType.IMGT_CDR3):
        identifier = uuid4().hex
        filename_base = filename_base if filename_base is not None else identifier

        data = make_airr_seq_set_object_from_sequences(sequences, region_type)

        data_filename = result_path / f"{filename_base}.tsv"
        bnp_write_to_file(data_filename, data)

        dynamic_fields = {f.name: f.type for f in fields(data)
                          if f not in [airr_f.name for airr_f in fields(AIRRSequenceSet)]}

        metadata_filename = result_path / f"{filename_base}.yaml"
        metadata = {} if metadata is None else metadata
        metadata['type_dict_dynamic_fields'] = dynamic_fields
        write_yaml(metadata_filename, metadata)

        repertoire = Repertoire(data_filename, metadata_filename, metadata, identifier, dynamic_fields=dynamic_fields)
        repertoire.element_count = len(data)

        return repertoire

    @property
    def bnp_dataclass(self):
        if self.dynamic_fields:
            return extend_dataclass_with_dynamic_fields(AIRRSequenceSet, list(self.dynamic_fields.items()))
        else:
            return AIRRSequenceSet

    @property
    def buffer_type(self):
        return get_bufferclass_for_datatype(self.bnp_dataclass, delimiter='\t', has_header=True)

    @property
    def data(self) -> AIRRSequenceSet:
        return bnp_read_from_file(self.data_filename, self.buffer_type, self.bnp_dataclass)

    def sequences(self, region_type: RegionType = RegionType.IMGT_CDR3) -> List[ReceptorSequence]:
        return make_sequences_from_data(self.data, self.dynamic_fields, region_type)

    def receptors(self, region_type: RegionType) -> List[Receptor]:
        return make_receptors_from_data(self.data, self.dynamic_fields, f'Repertoire {self.identifier}',
                                        region_type)

    def get_element_count(self):
        if not self.element_count:
            self.element_count = len(self.data)
        return self.element_count

    def get_locus(self):
        return sorted(set(self.data.locus.tolist()))

def build_dynamic_airr_sequence_set_dataclass(all_fields_dict: Dict[str, Any]):
    sequence_field_names = {seq_field.name: seq_field.type for seq_field in fields(AIRRSequenceSet)}
    types = {}

    for key, values in all_fields_dict.items():
        if key not in sequence_field_names:
            field_type = get_field_type_from_values(values)
            types[key] = field_type

    if types:
        dc = extend_dataclass_with_dynamic_fields(AIRRSequenceSet, list(types.items()))
    else:
        dc = AIRRSequenceSet
    return dc, types


def make_sequences_from_data(data, dynamic_fields: dict, region_type: RegionType = RegionType.IMGT_CDR3):
    seqs = []
    dynamic_fields_internal = dynamic_fields if dynamic_fields is not None else {}
    for el in data.to_iter():
        seq, seq_aa = get_sequence_value(el, region_type)
        seqs.append(ReceptorSequence(sequence_id=el.sequence_id, sequence=seq, sequence_aa=seq_aa,
                                     productive=el.productive, vj_in_frame=el.vj_in_frame, stop_codon=el.stop_codon,
                                     locus=el.locus, v_call=el.v_call,
                                     d_call=el.d_call if hasattr(el, 'd_call') else '',
                                     c_call=getattr(el, 'c_call', ''),
                                     j_call=el.j_call, duplicate_count=el.duplicate_count,
                                     metadata={dynamic_field: getattr(el, dynamic_field)
                                               for dynamic_field in dynamic_fields_internal.keys()}))
    return seqs


def make_receptors_from_data(data: AIRRSequenceSet, dynamic_fields: dict, location,
                             region_type: RegionType = RegionType.IMGT_CDR3):
    df = data.topandas()
    receptors = []
    for name, group in df.groupby('cell_id'):
        assert group.shape[0] == 2, \
            (f"{location}: receptor objects cannot be created from the data file, there are "
             f"{group.shape[0]} sequences with cell id {group['cell_id'].unique()[0]}.")
        sorted_group = group.sort_values(by='locus')
        seqs = [ReceptorSequence(sequence_id=el.sequence_id, sequence=getattr(el, region_type.value),
                                 sequence_aa=getattr(el, region_type.value + "_aa"),
                                 productive=el.productive, vj_in_frame=el.vj_in_frame, stop_codon=el.stop_codon,
                                 locus=el.locus, locus_species=el.locus_species, v_call=el.v_call,
                                 d_call=el.d_call, j_call=el.j_call, c_call=el.c_call,
                                 duplicate_count=el.duplicate_count,
                                 metadata={dynamic_field: getattr(el, dynamic_field)
                                           for dynamic_field in dynamic_fields.keys()})
                for index, el in sorted_group.iterrows()]

        if 'receptor_id' in group and group['receptor_id'].nunique() == 1:
            receptor_id = group['receptor_id'].unique()[0]
        else:
            receptor_id = uuid4().hex

        receptor = Receptor(chain_pair=ChainPair.get_chain_pair([Chain.get_chain(locus) for locus in group.locus]),
                            chain_1=seqs[0], chain_2=seqs[1], cell_id=group['cell_id'].unique()[0],
                            receptor_id=receptor_id,
                            metadata={key: list({seqs[0].metadata[key], seqs[1].metadata[key]})
                                      for key in dynamic_fields.keys()})
        receptors.append(receptor)

    return receptors


def make_airr_seq_set_object_from_sequences(sequences: List[ReceptorSequence],
                                            region_type: RegionType = RegionType.IMGT_CDR3):
    seq_fields = {key: [] for key in vars(ReceptorSequence()).keys()
                  if key not in ['metadata', 'sequence', 'sequence_aa']}
    seq_content_fields = {region_type.value: [], region_type.value + "_aa": []}

    dynamic_fields = {}
    for index, sequence in enumerate(sequences):
        seq_content_fields[region_type.value].append(sequence.sequence)
        seq_content_fields[region_type.value + "_aa"].append(sequence.sequence_aa)

        for key in seq_fields.keys():
            seq_fields[key].append(getattr(sequence, key))
        if sequence.metadata:
            for key in sequence.metadata.keys():
                if key in dynamic_fields:
                    dynamic_fields[key].append(sequence.metadata[key])
                elif index == 0:
                    dynamic_fields[key] = [sequence.metadata[key]]
                else:
                    raise RuntimeError("Sequences cannot be converted to AIRRSequenceSet because the metadata "
                                       "attributes between sequences differ.")

    all_fields_dict = {**seq_fields, **dynamic_fields, **seq_content_fields}
    return build_dynamic_bnp_dataclass_obj(all_fields_dict)[0]


def get_sequence_value(el: AIRRSequenceSet, region_type: RegionType = RegionType.IMGT_CDR3):
    if region_type == RegionType.FULL_SEQUENCE:
        return el.sequence, el.sequence_aa
    else:
        return getattr(el, region_type.value, ''), getattr(el, region_type.value + "_aa", '')



INTERNAL_EVAL_METRICS = ['calinski_harabasz_score', 'davies_bouldin_score', 'silhouette_score']
EXTERNAL_EVAL_METRICS = ['rand_score', 'adjusted_rand_score', 'adjusted_mutual_info_score', 'completeness_score',
                         'fowlkes_mallows_score', 'homogeneity_score', 'mutual_info_score',
                         'normalized_mutual_info_score', 'v_measure_score']


def is_internal(metric: str):
    return metric in INTERNAL_EVAL_METRICS


def is_external(metric: str):
    return metric in EXTERNAL_EVAL_METRICS


def is_valid_metric(metric: str):
    return metric in INTERNAL_EVAL_METRICS or metric in EXTERNAL_EVAL_METRICS



from collections import Counter

from immuneML.data_model.SequenceParams import Chain
from immuneML.data_model.datasets.ElementDataset import ReceptorDataset
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.kmer_frequency.KmerFrequencyEncoder import KmerFrequencyEncoder


class KmerFreqReceptorEncoder(KmerFrequencyEncoder):

    def _encode_locus(self, dataset):
        return True

    def _encode_new_dataset(self, dataset, params: EncoderParams):
        encoded_data = self._encode_data(dataset, params)

        encoded_dataset = dataset.clone()
        encoded_dataset.encoded_data = encoded_data

        return encoded_dataset

    def _encode_examples(self, dataset: ReceptorDataset, params: EncoderParams):
        encoded_receptors_counts = []
        receptor_ids = []
        label_config = params.label_config
        labels = {label: [] for label in label_config.get_labels_by_name()} if params.encode_labels else None

        params.region_type = self.region_type
        encode_locus = self._encode_locus(dataset)

        sequence_encoder = self._prepare_sequence_encoder()
        for receptor in dataset.get_data(region_type=self.region_type):
            chains = [Chain.get_chain(chain).name.lower() for chain in receptor.chain_pair.value]
            counts = {chain: Counter() for chain in chains}
            for chain in chains:
                counts[chain] = self._encode_sequence(getattr(receptor, chain), params, sequence_encoder,
                                                      counts[chain], encode_locus)
            encoded_receptors_counts.append(counts[chains[0]] + counts[chains[1]])
            receptor_ids.append(receptor.receptor_id)

            if params.encode_labels:
                for label_name in label_config.get_labels_by_name():
                    label = receptor.metadata[label_name]
                    labels[label_name].append(label)

        return encoded_receptors_counts, receptor_ids, labels



from dataclasses import dataclass, field
from pathlib import Path
from typing import List

from immuneML.IO.dataset_export.DataExporter import DataExporter
from immuneML.data_model.datasets.Dataset import Dataset


@dataclass
class SubsamplingState:

    dataset: Dataset
    subsampled_dataset_sizes: List[int] = field(default_factory=list)
    subsampled_repertoire_size: int = None
    result_path: Path = None
    name: str = None
    subsampled_datasets: List[Dataset] = field(default_factory=list)
    subsampled_dataset_paths: dict = field(default_factory=dict)



from pathlib import Path
from scripts.DocumentatonFormat import DocumentationFormat


def update_docs_per_mapping(docs: str, mapping: dict):
    for key in mapping:
        docs = docs.replace(key, mapping[key])
    return docs


def write_class_docs(doc_format: DocumentationFormat, file):
    title = f"\n{doc_format.cls_name}\n{doc_format.level_heading}\n\n" if doc_format.cls_name != "" else "\n\n"
    file.writelines(title)
    if hasattr(doc_format.cls, "get_documentation"):
        file.writelines("\n".join([el.replace('    ', '', 1) if el.startswith('    ') else el for el in doc_format.cls.get_documentation().split("\n")]))
        file.writelines('\n')
    elif doc_format.cls.__doc__ is not None:
        file.writelines("\n".join([el.replace('    ', '', 1) if el.startswith('    ') else el for el in doc_format.cls.__doc__.split("\n")]))
        file.writelines('\n')


def make_docs(path: Path, classes, filename, drop_name_part, file_open_mode="w", format_level=DocumentationFormat.LEVELS[1]):
    classes.sort(key=lambda cls: cls.__name__)
    classes_to_document = [DocumentationFormat(cls, cls.__name__.replace(drop_name_part, ""), format_level)
                           for cls in classes]

    file_path = path / filename
    with file_path.open(file_open_mode) as file:
        for doc_format in classes_to_document:
            write_class_docs(doc_format, file)



from immuneML.IO.dataset_import.DataImport import DataImport
from immuneML.data_model.datasets.ElementDataset import ReceptorDataset
from immuneML.simulation.dataset_generation.RandomDatasetGenerator import RandomDatasetGenerator
from immuneML.util.ParameterValidator import ParameterValidator


class RandomReceptorDatasetImport(DataImport):
    """
    Returns a ReceptorDataset consisting of randomly generated sequences, which can be used for benchmarking purposes.
    The sequences consist of uniformly chosen amino acids or nucleotides.


    **Specification arguments:**

    - receptor_count (int): The number of receptors the ReceptorDataset should contain.

    - chain_1_length_probabilities (dict): A mapping where the keys correspond to different sequence lengths for chain
      1, and the values are the probabilities for choosing each sequence length. For example, to create a random
      ReceptorDataset where 40% of the sequences for chain 1 would be of length 10, and 60% of the sequences would
      have length 12, this mapping would need to be specified:

    .. indent with spaces
    .. code-block:: yaml

        10: 0.4
        12: 0.6

    - chain_2_length_probabilities (dict): Same as chain_1_length_probabilities, but for chain 2.

    - labels (dict): A mapping that specifies randomly chosen labels to be assigned to the receptors. One or multiple
      labels can be specified here. The keys of this mapping are the labels, and the values consist of another mapping
      between label classes and their probabilities. For example, to create a random ReceptorDataset with the label
      cmv_epitope where 70% of the receptors has class binding and the remaining 30% has class not_binding, the
      following mapping should be specified:

    .. indent with spaces
    .. code-block:: yaml

        cmv_epitope:
            binding: 0.7
            not_binding: 0.3


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            datasets:
                my_random_dataset:
                    format: RandomReceptorDataset
                    params:
                        receptor_count: 100 # number of random receptors to generate
                        chain_1_length_probabilities:
                            14: 0.8 # 80% of all generated sequences for all receptors (for chain 1) will have length 14
                            15: 0.2 # 20% of all generated sequences across all receptors (for chain 1) will have length 15
                        chain_2_length_probabilities:
                            14: 0.8 # 80% of all generated sequences for all receptors (for chain 2) will have length 14
                            15: 0.2 # 20% of all generated sequences across all receptors (for chain 2) will have length 15
                        labels:
                            epitope1: # label name
                                True: 0.5 # 50% of the receptors will have class True
                                False: 0.5 # 50% of the receptors will have class False
                            epitope2: # next label with classes that will be assigned to receptors independently of the previous label or other parameters
                                1: 0.3 # 30% of the generated receptors will have class 1
                                0: 0.7 # 70% of the generated receptors will have class 0

    """
    def __init__(self, params: dict, dataset_name: str):
        super().__init__({}, dataset_name)
        self.params = params
        self.dataset_name = dataset_name

    def import_dataset(self) -> ReceptorDataset:
        """
        Returns randomly generated receptor dataset according to the parameters;

        **YAML specification:**

            result_path: path/where/to/store/results/
            receptor_count: 100 # number of random receptors to generate
            chain_1_length_probabilities:
                14: 0.8 # 80% of all generated sequences for all receptors (for chain 1) will have length 14
                15: 0.2 # 20% of all generated sequences across all receptors (for chain 1) will have length 15
            chain_2_length_probabilities:
                14: 0.8 # 80% of all generated sequences for all receptors (for chain 2) will have length 14
                15: 0.2 # 20% of all generated sequences across all receptors (for chain 2) will have length 15
            labels:
                epitope1: # label name
                    True: 0.5 # 50% of the receptors will have class True
                    False: 0.5 # 50% of the receptors will have class False
                epitope2: # next label with classes that will be assigned to receptors independently of the previous label or other parameters
                    1: 0.3 # 30% of the generated receptors will have class 1
                    0: 0.7 # 70% of the generated receptors will have class 0

        """
        valid_keys = ["receptor_count", "chain_1_length_probabilities", "chain_2_length_probabilities", "labels", "result_path"]
        ParameterValidator.assert_all_in_valid_list(list(self.params.keys()), valid_keys, "RandomReceptorDatasetImport", "params")

        return RandomDatasetGenerator.generate_receptor_dataset(receptor_count=self.params["receptor_count"],
                                                                chain_1_length_probabilities=self.params["chain_1_length_probabilities"],
                                                                chain_2_length_probabilities=self.params["chain_2_length_probabilities"],
                                                                labels=self.params["labels"],
                                                                path=self.params["result_path"],
                                                                name=self.dataset_name)



from collections import Counter
from pathlib import Path

import pandas as pd
import plotly.express as px

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.data_model.datasets.ElementDataset import ReceptorDataset, SequenceDataset
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.ReportUtil import ReportUtil
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.util.PathBuilder import PathBuilder


class SequenceCountDistribution(DataReport):
    """
    Generates a histogram of the duplicate counts of the sequences in a dataset.


    **Specification arguments:**

    - split_by_label (bool): Whether to split the plots by a label. If set to true, the Dataset must either contain a single label, or alternatively the label of interest can be specified under 'label'. By default, split_by_label is False.

    - label (str): Optional label for separating the results by color/creating separate plots. Note that this should the name of a valid dataset label.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        my_sld_report:
            SequenceCountDistribution:
                label: disease

    """

    @classmethod
    def build_object(cls, **kwargs):
        location = SequenceCountDistribution.__name__

        ReportUtil.update_split_by_label_kwargs(kwargs, location)

        return SequenceCountDistribution(**kwargs)

    def __init__(self, dataset: Dataset = None, result_path: Path = None, number_of_processes: int = 1,
                 split_by_label: bool = None, label: str = None, name: str = None):
        super().__init__(dataset=dataset, result_path=result_path, number_of_processes=number_of_processes, name=name)
        self.split_by_label = split_by_label
        self.label_name = label

    def _set_label_name(self):
        if self.split_by_label:
            if self.label_name is None:
                self.label_name = list(self.dataset.get_label_names())[0]
        else:
            self.label_name = None

    def check_prerequisites(self):
        return True

    def _generate(self) -> ReportResult:
        self._set_label_name()

        df = self._get_sequence_counts_df()
        PathBuilder.build(self.result_path)

        output_table = self._write_output_table(df, self.result_path / "sequence_count_distribution.tsv",
                                                name="Duplicate counts of sequences in the dataset")

        report_output_fig = self._safe_plot(df=df, output_written=False)
        output_figures = None if report_output_fig is None else [report_output_fig]
        return ReportResult(name=self.name,
                            info="The sequence count distribution of the dataset.",
                            output_figures=output_figures, output_tables=[output_table])

    def _get_sequence_counts_df(self):
        if isinstance(self.dataset, RepertoireDataset):
            return self._get_repertoire_df()
        elif isinstance(self.dataset, ReceptorDataset) or isinstance(self.dataset, SequenceDataset):
            return self._get_sequence_receptor_df()

    def _get_repertoire_df(self):
        sequence_counts = Counter()

        for repertoire in self.dataset.get_data():
            if self.split_by_label:
                label_class = repertoire.metadata[self.label_name]
            else:
                label_class = None

            repertoire_counter = Counter(repertoire.data.duplicate_count)
            sequence_counts += Counter({(key, label_class): value for key, value in repertoire_counter.items()})

        df = pd.DataFrame({"n_observations": list(sequence_counts.values()),
                           "duplicate_count": [key[0] for key in sequence_counts.keys()]})

        if self.split_by_label:
            df[self.label_name] = [key[1] for key in sequence_counts.keys()]

        return df

    def _get_sequence_receptor_df(self):

        data = self.dataset.data

        try:
            counts = data.duplicate_count
        except AttributeError as e:
            raise AttributeError(
                f"{SequenceCountDistribution.__name__}: SequenceDataset does not contain attribute 'duplicate_count'. "
                f"This report can only be run when sequence counts are available.")

        chains = data.locus.tolist()

        if self.split_by_label:
            label_classes = getattr(data, self.label_name).tolist()
            counter = Counter(zip(counts, chains, label_classes))
        else:
            counter = Counter(zip(counts, chains))

        df = pd.DataFrame({"duplicate_count": [key[0] for key in counter.keys()],
                           "locus": [key[1] for key in counter.keys()],
                           "n_observations": counter.values()})

        if self.split_by_label:
            df[self.label_name] = [key[2] for key in counter.keys()]

        return df

    def _plot(self, df: pd.DataFrame) -> ReportOutput:
        figure = px.bar(df, x="duplicate_count", y="n_observations", barmode="group",
                        color=self.label_name if self.split_by_label else None,
                        facet_col="locus" if "locus" in df.columns and len(set(df["locus"])) > 1 else None,
                        color_discrete_sequence=px.colors.diverging.Tealrose,
                        labels={"n_observations": "Number of observations",
                                "duplicate_count": "Sequence duplicate count"})
        figure.update_layout(template="plotly_white")
        PathBuilder.build(self.result_path)

        file_path = self.result_path / "sequence_count_distribution.html"
        file_path = PlotlyUtil.write_image_to_file(figure, file_path)
        return ReportOutput(path=file_path, name="Sequence duplicate count distribution")



import logging
import shutil
from pathlib import Path

import bionumpy
import pandas as pd

from immuneML.IO.dataset_export.DataExporter import DataExporter
from immuneML.data_model.AIRRSequenceSet import AIRRSequenceSet
from immuneML.data_model.bnp_util import read_yaml, write_yaml, write_dataset_yaml
from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.data_model.datasets.ElementDataset import ElementDataset
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.util.PathBuilder import PathBuilder


class AIRRExporter(DataExporter):
    """
    Exports a RepertoireDataset of Repertoires in AIRR format.

    Things to note:

    - one filename_prefix is given, which is combined with the Repertoire identifiers for the filenames, to create one file per Repertoire
    - 'counts' is written into the field 'duplicate_counts'
    - 'sequence_identifiers' is written both into the fields 'sequence_id' and 'rearrangement_id'


    """

    @staticmethod
    def export(dataset: Dataset, path: Path, number_of_processes: int = 1, omit_columns: list = None):
        PathBuilder.build(path)

        try:

            if isinstance(dataset, RepertoireDataset):
                AIRRExporter.export_repertoire_dataset(dataset, path, omit_columns=omit_columns)
            elif isinstance(dataset, ElementDataset):
                AIRRExporter.process_and_store_data_file(dataset.filename, path / dataset.filename.name)
                AIRRExporter.process_and_store_dataset_file(dataset.dataset_file, path / dataset.dataset_file.name)

        except shutil.SameFileError as e:
            logging.warning(f"AIRRExporter: target and input path are the same. Skipping the copy operation...")

        # TODO: add here export of full sequence if possible

    @staticmethod
    def export_repertoire_dataset(dataset: RepertoireDataset, path: Path, omit_columns: list = None):
        repertoire_folder = "repertoires/"
        repertoire_path = PathBuilder.build(path / repertoire_folder)

        AIRRExporter.export_repertoire_files(dataset, repertoire_path)

        shutil.copyfile(dataset.metadata_file, path / f"{dataset.name}.csv")

        if dataset.dataset_file and dataset.dataset_file.is_file():
            dataset_yaml = read_yaml(dataset.dataset_file)
            dataset_yaml['metadata_file'] = path / f"{dataset.name}.csv"
        else:

            new_metadata_df = pd.read_csv(path / f"{dataset.name}.csv")
            labels = {label: new_metadata_df[label].unique().tolist()
                      for label in dataset.get_label_names(refresh=True)}

            dataset_yaml = RepertoireDataset.create_metadata_dict(metadata_file=path / f"{dataset.name}.csv",
                                                                  labels=labels,
                                                                  name=dataset.name,
                                                                  identifier=dataset.identifier)
        write_dataset_yaml(path / f"{dataset.name}.yaml", dataset_yaml)

    @staticmethod
    def export_repertoire_files(dataset: RepertoireDataset, repertoire_path: Path):
        for repertoire in dataset.repertoires:
            AIRRExporter.process_and_store_data_file(repertoire.data_filename,
                                                     repertoire_path / repertoire.data_filename.name)
            shutil.copyfile(repertoire.metadata_filename, repertoire_path / repertoire.metadata_filename.name)

    @staticmethod
    def process_and_store_dataset_file(input_filename, output_filename):
        metadata = read_yaml(input_filename)
        metadata['filename'] = str(Path(metadata['filename']).name)
        write_yaml(output_filename, metadata)

    @staticmethod
    def process_and_store_data_file(input_filename, output_filename):
        df = pd.read_csv(input_filename, sep='\t', keep_default_na=False,
                         dtype={key: key_type if not isinstance(key_type, bionumpy.encodings.Encoding) else str
                                for key, key_type in AIRRSequenceSet.get_field_type_dict().items()})
        df = AIRRExporter.add_cdr3_from_junction(df)
        df.to_csv(output_filename, sep='\t', index=False)

    @staticmethod
    def add_cdr3_from_junction(df: pd.DataFrame) -> pd.DataFrame:
        if 'junction' in df.columns and 'cdr3' in df.columns and any(df.cdr3.eq('')):
            missing_cdr3 = df.cdr3.eq('')
            df.loc[missing_cdr3, 'cdr3'] = df.loc[missing_cdr3, 'junction'].str[3:-3]
        if 'junction_aa' in df.columns and 'cdr3_aa' in df.columns and any(df.cdr3_aa.eq('')):
            missing_cdr3_aa = df.cdr3_aa.eq('')
            df.loc[missing_cdr3_aa, 'cdr3_aa'] = df.loc[missing_cdr3_aa, 'junction_aa'].str[1:-1]
        return df



import logging

import random
import numpy as np
from sklearn.preprocessing import label_binarize

from immuneML.data_model.EncodedData import EncodedData
from immuneML.environment.Constants import Constants
from immuneML.environment.Label import Label


class Util:

    @staticmethod
    def map_to_old_class_values(y, class_mapping: dict):
        try:
            old_class_type = np.array(list(class_mapping.values())).dtype
            mapped_y = np.copy(y).astype(object)
            for i in range(mapped_y.shape[0]):
                mapped_y[i] = class_mapping[y[i]]
            return mapped_y.astype(old_class_type)
        except Exception as e:
            logging.exception(
                "MLMethod util: error occurred when predicting the class assignment due to mismatch of class types.\n"
                f"Classes: {y}\nMapping:{class_mapping}")
            raise e

    @staticmethod
    def map_to_new_class_values(y, class_mapping: dict):
        try:
            mapped_y = np.copy(y).astype(object)
            switched_mapping = {value: key for key, value in class_mapping.items()}
            new_class_type = np.array(list(switched_mapping.values())).dtype
            for i in range(mapped_y.shape[0]):
                mapped_y[i] = switched_mapping[y[i]]
            return mapped_y.astype(new_class_type)
        except Exception as e:
            logging.exception(f"MLMethod util: error occurred when fitting the model due to mismatch of class types.\n"
                              f"Classes: {y}\nMapping:{class_mapping}")
            raise

    @staticmethod
    def make_class_mapping(y, label: Label) -> dict:
        """Creates a class mapping from a list of classes which can be strings, numbers of booleans; maps to same
        name in multi-class settings"""
        classes = label.values
        if len(classes) == 2:
            return Util.make_binary_class_mapping(y, label)
        else:
            return {cls: cls for cls in classes}

    @staticmethod
    def make_binary_class_mapping(y, label: Label) -> dict:
        """
        Creates binary class mapping from a list of classes which can be strings, numbers or boolean values

        Arguments:

            y: list of classes per example, as supplied to fit() method of the classifier; it should include all
            classes that will appear in the data

            label: label object along with information on the positive label

        Returns:
             mapping dictionary where 0 and 1 are always the keys and the values are original class names which were
             mapped for these values

        """
        unique_values = sorted(set(label.values))
        assert len(unique_values) == 2, f"MLMethod: there has two be exactly two classes to use this classifier," \
                                        f" instead got {str(unique_values)[1:-1]}. For multi-class classification, " \
                                        f"consider some of the other classifiers."

        if label.positive_class is None:
            return {0: unique_values[0], 1: unique_values[1]}
        else:
            assert label.positive_class in unique_values, \
                (f"MLMethod: the specified positive class '{label.positive_class}' does not occur in the list of "
                 f"available classes: {str(unique_values)[1:-1]}.")
            unique_values.remove(label.positive_class)
            return {0: unique_values[0], 1: label.positive_class}

    @staticmethod
    def binarize_label_classes(true_y, predicted_y, classes):
        """
        Binarizes the predictions in place using scikit-learn's label_binarize() method

        Necessary for some sklearn metrics, like roc_auc_score
        """
        if hasattr(true_y, 'dtype') and (true_y.dtype.type is np.str_ or true_y.dtype.type is np.object_) \
                or isinstance(true_y, list) and any(isinstance(item, str) for item in true_y):
            true_y = label_binarize(true_y, classes=classes)
            predicted_y = label_binarize(predicted_y, classes=classes)

        return true_y, predicted_y

    @staticmethod
    def setup_pytorch(number_of_threads, random_seed, pytorch_device_name=None):
        import torch
        torch.set_num_threads(number_of_threads)
        torch.manual_seed(random_seed)
        if pytorch_device_name is not None:
            torch.device(pytorch_device_name)

    @staticmethod
    def get_immuneML_version():
        return 'immuneML ' + Constants.VERSION

    @staticmethod
    def get_train_val_indices(n_examples, training_percentage, random_seed=None):
        indices = list(range(n_examples))

        random.seed(random_seed)
        random.shuffle(indices)
        random.seed(None)

        limit = int(n_examples * training_percentage)
        train_indices = indices[:limit]
        val_indices = indices[limit:]

        return train_indices, val_indices

    @staticmethod
    def subset_encoded_data(encoded_data: EncodedData, indices):
        return EncodedData(examples=encoded_data.examples[indices],
                           labels={label_name: [encoded_data.labels[label_name][i] for i in indices]
                                   for label_name in encoded_data.labels.keys()},
                           example_ids=[encoded_data.example_ids[i] for i in indices],
                           example_weights=[encoded_data.example_weights[i] for i in
                                            indices] if encoded_data.example_weights is not None else None,
                           feature_names=encoded_data.feature_names,
                           feature_annotations=encoded_data.feature_annotations,
                           encoding=encoded_data.encoding,
                           info=encoded_data.info)



from abc import ABC
from pathlib import Path

from immuneML.reports.Report import Report
from immuneML.workflows.instructions.clustering.clustering_run_model import ClusteringItem


class ClusteringMethodReport(Report, ABC):
    """
    Clustering method reports show some features or statistics about the clustering method.
    """

    DOCS_TITLE = "Clustering method reports"

    def __init__(self, result_path: Path = None, name: str = None, clustering_item: ClusteringItem = None):
        super().__init__(name=name, result_path=result_path)
        self.item = clustering_item



import logging
from dataclasses import make_dataclass, fields as get_fields
from itertools import chain
from pathlib import Path
from typing import List, Dict, Union

import bionumpy as bnp
import numpy as np
import pandas as pd
from bionumpy import DNAEncoding, get_motif_scores, AminoAcidEncoding
from bionumpy.bnpdataclass import bnpdataclass, BNPDataClass
from bionumpy.encodings import BaseEncoding
from bionumpy.io import delimited_buffers
from bionumpy.sequence.string_matcher import RegexMatcher, StringMatcher
from npstructures import RaggedArray
from scipy.stats import zipf

from immuneML import Constants
from immuneML.data_model.SequenceParams import RegionType, Chain
from immuneML.data_model.SequenceSet import Repertoire
from immuneML.data_model.bnp_util import make_full_airr_seq_set_df
from immuneML.environment.SequenceType import SequenceType
from immuneML.ml_methods.generative_models.BackgroundSequences import BackgroundSequences
from immuneML.simulation.SimConfigItem import SimConfigItem
from immuneML.simulation.implants.LigoPWM import LigoPWM
from immuneML.simulation.implants.MotifInstance import MotifInstance, MotifInstanceGroup
from immuneML.simulation.implants.Signal import Signal, SignalPair
from immuneML.simulation.util.bnp_util import merge_dataclass_objects, pad_ragged_array
from immuneML.util.PathBuilder import PathBuilder
from immuneML.util.PositionHelper import PositionHelper


def get_signal_sequence_count(repertoire_count: int, signal_proportion: float,
                              receptors_in_repertoire_count: int) -> int:
    return round(receptors_in_repertoire_count * signal_proportion) * repertoire_count


def get_sequence_per_signal_count(sim_item: SimConfigItem) -> dict:
    if sim_item.receptors_in_repertoire_count:
        sequence_count = sim_item.receptors_in_repertoire_count * sim_item.number_of_examples
        seq_with_signal_count = {signal.id: get_signal_sequence_count(sim_item.number_of_examples, proportion,
                                                                      sim_item.receptors_in_repertoire_count)
                                 for signal, proportion in sim_item.signal_proportions.items()}
        seq_without_signal_count = {'no_signal': sequence_count - sum(seq_with_signal_count.values())}

        return {**seq_with_signal_count, **seq_without_signal_count}
    elif len(sim_item.signals) == 1:
        return {sim_item.signals[0].id: sim_item.number_of_examples, 'no_signal': 0}
    elif len(sim_item.signals) == 0:
        return {'no_signal': sim_item.number_of_examples}
    else:
        raise NotImplementedError


def get_bnp_data(sequence_path, bnp_data_class):
    if sequence_path.is_file():
        buff_type = delimited_buffers.get_bufferclass_for_datatype(bnp_data_class, delimiter='\t', has_header=True)

        with bnp.open(sequence_path, buffer_type=buff_type) as file:
            data = file.read()

        return data


def construct_sequence_metadata_object(sequence, metadata: dict, custom_params, immune_events: dict, locus: Chain) \
        -> dict:
    custom = {}

    for param in custom_params:
        key, key_type = param[0], param[1]
        if any(str_key in key for str_key in ['position', 'signals_aggregated', 'original_sequence']):
            custom[key] = getattr(sequence, key).to_string()
        else:
            custom[key] = getattr(sequence, key).item()

    return {**metadata, **custom, **immune_events}


def write_bnp_data(path: Path, data, append_if_exists: bool = True):
    if len(data) > 0:
        buff_type = delimited_buffers.get_bufferclass_for_datatype(type(data), delimiter="\t", has_header=True)

        if path.is_file() and append_if_exists:
            with bnp.open(path, buffer_type=buff_type, mode='a') as file:
                file.write(data)
        elif not path.is_file():
            with bnp.open(path, buffer_type=buff_type, mode='w') as file:
                file.write(data)
        else:
            raise RuntimeError(
                f"Tried writing to {path}, but it already exists and append_if_exists parameter is set to False.")


def make_sequence_paths(path: Path, signals: List[Signal]) -> Dict[str, Path]:
    tmp_path = PathBuilder.build(path / 'processed_sequences')
    sequence_paths = {signal.id: tmp_path / f'{signal.id}.tsv' for signal in signals}
    sequence_paths['no_signal'] = tmp_path / 'no_signal.tsv'
    return sequence_paths


def get_allowed_positions(signal: Signal, sequence_array: RaggedArray, region_type: RegionType):
    sequence_lengths = sequence_array.lengths
    if bool(signal.sequence_position_weights):
        allowed_positions = RaggedArray(
            [PositionHelper.get_allowed_positions_for_annotation(seq_len, region_type, signal.sequence_position_weights)
             for seq_len in sequence_lengths])
    else:
        allowed_positions = None

    return allowed_positions


def get_region_type(sequences) -> RegionType:
    if hasattr(sequences, "region_type") and \
            np.all([el.to_string() == getattr(sequences, 'region_type')[0].to_string() for el in
                    getattr(sequences, 'region_type')]):
        return RegionType[getattr(sequences, 'region_type')[0].to_string()]
    else:
        raise RuntimeError(f"The region types could not be obtained.")


def annotate_sequences(sequences: BackgroundSequences, is_amino_acid: bool, all_signals: list, annotated_dc, sim_item_name: str = None,
                       region_type: RegionType = RegionType.IMGT_CDR3):
    encoding = AminoAcidEncoding if is_amino_acid else DNAEncoding
    sequence_array = sequences.sequence_aa if is_amino_acid else sequences.sequence

    signal_matrix = np.zeros((len(sequence_array), len(all_signals)))
    signal_positions = {}

    for index, signal in enumerate(all_signals):
        _annotate_with_signal(sequences, sequence_array, is_amino_acid, encoding, signal_matrix, signal, index,
                              signal_positions, sim_item_name, region_type)

    signal_matrix = make_bnp_annotated_sequences(sequences, annotated_dc, all_signals, signal_matrix, signal_positions)

    logging.info(f"Annotated {len(sequences)} sequences with signal information.")

    return signal_matrix


def _annotate_with_signal(sequences, sequence_array, is_amino_acid, encoding, signal_matrix, signal, signal_index,
                          signal_positions, sim_item_name: str = None, region_type: RegionType = RegionType.IMGT_CDR3):
    if signal.motifs is not None:
        return _annotate_with_signal_motifs(sequences, sequence_array, is_amino_acid, encoding, signal_matrix, signal,
                                            signal_index, signal_positions, sim_item_name, region_type)
    else:
        return _annotate_with_signal_func(sequences, sequence_array, is_amino_acid, encoding, signal_matrix, signal,
                                          signal_index, signal_positions, region_type)


def _annotate_with_signal_func(sequences, sequence_array, is_amino_acid, encoding, signal_matrix, signal, signal_index,
                               signal_positions, region_type: RegionType):
    signal_matrix[:, signal_index] = [signal.is_present_custom_func(seq.sequence_aa.to_string(),
                                                                    seq.sequence.to_string(),
                                                                    seq.v_call.to_string(),
                                                                    seq.j_call.to_string(),
                                                                    region_type.name)
                                      for seq in sequences]
    signal_positions[f'{signal.id}_positions'] = ["" for _ in range(len(sequences))]


def _annotate_with_signal_motifs(sequences, sequence_array, is_amino_acid, encoding, signal_matrix, signal,
                                 signal_index, signal_positions, sim_item_name: str = None,
                                 region_type: RegionType = RegionType.IMGT_CDR3):
    signal_pos_col = None
    allowed_positions = get_allowed_positions(signal, sequence_array, region_type)
    matches_gene = match_genes(signal.v_call, sequences.v_call, signal.j_call, sequences.j_call)

    for motifs in signal.get_all_motif_instances(SequenceType.AMINO_ACID if is_amino_acid else SequenceType.NUCLEOTIDE):
        matches = None

        if isinstance(motifs, MotifInstanceGroup):
            matches = match_motif_group(motifs, encoding, sequence_array, matches_gene, matches)
        else:
            matches = match_motif_regexes(motifs, encoding, sequence_array, matches_gene, matches)

        if allowed_positions is not None:
            matches = np.logical_and(matches, allowed_positions)

        signal_pos_col = np.logical_or(signal_pos_col, matches) if signal_pos_col is not None else matches
        signal_matrix[:, signal_index] = np.logical_or(signal_matrix[:, signal_index],
                                                       np.logical_or.reduce(matches, axis=1))
        # TODO: we want to remove a receptor if it has multiple matches of the motif -> xor? or somehow removed?

    np_mask = RaggedArray(np.where(signal_pos_col.ravel(), "1", "0"), shape=signal_pos_col.shape)
    signal_positions[f'{signal.id}_positions'] = ['m' + "".join(np_mask[ind, :]) for ind in range(len(signal_pos_col))]


def match_motif_regexes(motifs, encoding, sequence_array, matches_gene, matches):
    matches_motif = None
    if isinstance(motifs, LigoPWM):
        matches_motif = match_motif(motifs, encoding, sequence_array)
    else:
        for motif in motifs:
            if matches_motif is None:
                matches_motif = match_motif(motif, encoding, sequence_array)
            else:
                matches_motif = np.logical_or(matches_motif, match_motif(motif, encoding, sequence_array))

    if matches is None:
        matches = np.logical_and(matches_motif, matches_gene)
    else:
        matches = np.logical_or(matches, np.logical_and(matches_motif, matches_gene))

    return matches


def match_motif_group(motif_group: list, encoding, sequence_array, matches_gene, matches):
    """Match if two motifs co-occur in the same sequence"""

    assert len(motif_group) == 2, len(motif_group)

    matches_motif_1 = match_motif_regexes(motif_group[0], encoding, sequence_array, matches_gene, None)
    matches_motif_2 = match_motif_regexes(motif_group[1], encoding, sequence_array, matches_gene, None)
    matches_motif = np.zeros_like(matches_motif_1)
    selection = np.logical_and(matches_motif_1.any(axis=1), matches_motif_2.any(axis=1))
    matches_motif[selection] = np.logical_or(matches_motif_1[selection], matches_motif_2[selection])

    return np.logical_or(matches_motif, matches) if matches is not None else matches_motif


def match_genes(v_call, v_call_array, j_call, j_call_array):
    if v_call is not None and v_call != "":
        matcher = StringMatcher(v_call, encoding=BaseEncoding)
        matches_gene = matcher.rolling_window(v_call_array, mode='same').any(axis=1).reshape(-1, 1)
    else:
        matches_gene = np.ones(len(v_call_array)).reshape(-1, 1)

    if j_call is not None and j_call != "":
        matcher = StringMatcher(j_call, encoding=BaseEncoding)
        matches_j = matcher.rolling_window(j_call_array, mode='same').any(axis=1).reshape(-1, 1)
        matches_gene = np.logical_and(matches_gene, matches_j)

    return matches_gene.astype(bool)


def match_motif(motif: Union[str, LigoPWM], encoding, sequence_array):
    if isinstance(motif, str):
        matcher = RegexMatcher(motif, encoding=encoding)
        matches = matcher.rolling_window(sequence_array, mode='same')
    else:
        matches = get_motif_scores(bnp.as_encoded_array(sequence_array.tolist(), motif.pwm_matrix._encoding),
                                   motif.pwm_matrix) > motif.threshold
        matches = pad_ragged_array(matches, sequence_array.shape, False)
    return matches


def filter_out_illegal_sequences(sequences, sim_item: SimConfigItem, all_signals: list, max_signals_per_sequence: int,
                                 max_motifs_per_sequence: int):
    if max_signals_per_sequence > 2 or max_motifs_per_sequence > 1:
        raise NotImplementedError
    elif max_signals_per_sequence == -1 or all_signals is None or len(all_signals) == 0:
        return sequences

    sim_signal_ids = list(
        set(chain.from_iterable([signal.id.split(Constants.SIGNAL_DELIMITER) for signal in sim_item.signals])))
    other_signals = [signal.id not in sim_signal_ids for signal in all_signals]
    signal_matrix = sequences.get_signal_matrix()
    legal_indices = np.logical_and(signal_matrix.sum(axis=1) <= max_signals_per_sequence,
                                   np.array(signal_matrix[:, other_signals] == 0).all(axis=1) if any(
                                       other_signals) else 1)

    legal_indices &= np.array(
        [(getattr(sequences, f'{s.id}_positions')[:, 1:] == '1').sum(axis=1) <= max_motifs_per_sequence for s in
         all_signals]).all(axis=0)

    return sequences[legal_indices]


def make_signal_metadata(sim_item, signals) -> Dict[str, bool]:
    metadata = {}
    for signal in sim_item.signals:
        if isinstance(signal, SignalPair):
            metadata[signal.signal1.id] = True if not sim_item.is_noise else False
            metadata[signal.signal2.id] = True if not sim_item.is_noise else False
        else:
            metadata[signal.id] = True if not sim_item.is_noise else False

    return {**metadata, **{signal.id: False for signal in signals if signal.id not in metadata}}


def make_repertoire_from_sequences(sequences: BNPDataClass, repertoires_path, sim_item: SimConfigItem,
                                   signals: List[Signal]) -> Repertoire:
    metadata = {**make_signal_metadata(sim_item, signals), **sim_item.immune_events, 'sim_item': sim_item.name}
    rep_data = prepare_data_for_repertoire_obj(sequences)
    return Repertoire.build(path=repertoires_path, metadata=metadata, **rep_data)


def prepare_data_for_repertoire_obj(sequences):
    df = sequences.topandas()
    df = prepare_data_for_airr_seq_set(df)

    return df.to_dict(orient='list')


def prepare_data_for_airr_seq_set(in_df: pd.DataFrame) -> pd.DataFrame:
    region_type = RegionType[in_df.region_type.unique()[0]].value

    df = in_df.rename(columns={
        'sequence_aa': region_type + "_aa",
        'sequence': region_type
    }).drop(columns=['frame_type', 'region_type'])

    if "junction" in df.columns and "cdr3" not in df.columns:
        df['cdr3'] = df.junction.str[3:-3]
    if "junction_aa" in df.columns and "cdr3_aa" not in df.columns:
        df['cdr3_aa'] = df.junction_aa.str[1:-1]

    df = make_full_airr_seq_set_df(df)

    return df


def make_bnp_annotated_sequences(sequences: BackgroundSequences, bnp_data_class, all_signals: list,
                                 signal_matrix: np.ndarray,
                                 signal_positions: dict):
    kwargs = {**{s.id: signal_matrix[:, ind].astype(int) for ind, s in enumerate(all_signals)},
              **{f"{s.id}_positions": bnp.as_encoded_array(signal_positions[f"{s.id}_positions"],
                                                           bnp.encodings.BaseEncoding) for ind, s in
                 enumerate(all_signals)},
              **{field.name: getattr(sequences, field.name) for field in get_fields(sequences)}}

    dc_fields = get_fields(bnp_data_class)
    if any([f'observed_{s.id}' in dc_fields for s in all_signals]):
        kwargs = {**kwargs,
                  **{f'observed_{s.id}': signal_matrix[:, ind].astype(int) for ind, s in enumerate(all_signals)}}

    kwargs['signals_aggregated'] = [s if s != "" else "no_signal" for s in
                                    ["_".join(
                                        s for index, s in enumerate([sig.id for sig in all_signals]) if el[index] == 1)
                                        for el in signal_matrix]]

    for field in dc_fields:
        if field.name not in kwargs:
            kwargs[field.name] = [field.default for _ in range(len(sequences))]

    return bnp_data_class(**kwargs)


def make_annotated_dataclass(annotation_fields: list, signals: list):
    functions = {
        "get_signal_matrix": lambda self: np.array(
            [getattr(self, signal.id) for signal in signals]).T if signals and len(signals) > 0 else None,
        "get_signal_names": lambda self: [signal.id for signal in signals]}

    fields = [f if len(f) != 3 else (f[0], f[1], f[2]) for f in annotation_fields]

    return bnpdataclass(
        make_dataclass("AnnotatedGenData", namespace=functions, bases=tuple([BackgroundSequences]), fields=fields))


def build_imgt_positions(sequence_length: int, motif_instance: MotifInstance, sequence_region_type):
    assert sequence_length >= len(motif_instance), \
        "The motif instance is longer than sequence length. Remove the receptor_sequence from the repertoire or reduce max gap length " \
        "to be able to proceed."

    if sequence_region_type.to_string() == RegionType.IMGT_JUNCTION.name:
        return PositionHelper.gen_imgt_positions_from_junction_length(sequence_length)
    elif sequence_region_type.to_string() == RegionType.IMGT_CDR3.name:
        return PositionHelper.gen_imgt_positions_from_cdr3_length(sequence_length)
    else:
        raise NotImplementedError(
            f"IMGT positions here are defined only for CDR3 and JUNCTION region types, got {sequence_region_type}")


def choose_implant_position(imgt_positions, position_weights):
    imgt_implant_position = np.random.choice(list(position_weights.keys()), size=1, p=list(position_weights.values()))
    position = np.where(imgt_positions == imgt_implant_position)[0][0]
    return position


def check_iteration_progress(iteration: int, max_iterations: int):
    if iteration == round(max_iterations * 0.75):
        logging.warning(
            f"Iteration {iteration} out of {max_iterations} max iterations reached during simulation.")


def check_sequence_count(sim_item, sequences: BackgroundSequences):
    assert len(sequences) == sim_item.receptors_in_repertoire_count, \
        f"Error when simulating repertoire, needed {sim_item.receptors_in_repertoire_count} sequences, " \
        f"but got {len(sequences)}."


def update_seqs_without_signal(max_count, annotated_sequences, seqs_no_signal_path: Path):
    if max_count > 0:
        signal_matrix = annotated_sequences.get_signal_matrix()
        selection = signal_matrix.sum(axis=1) == 0 if signal_matrix is not None else np.ones(len(annotated_sequences),
                                                                                             dtype=bool)
        data_to_write = annotated_sequences[selection][:max_count]
        if len(data_to_write) > 0:
            write_bnp_data(data=data_to_write, path=seqs_no_signal_path)
        return max_count - len(data_to_write)
    else:
        return max_count


def update_seqs_with_signal(max_counts: dict, annotated_sequences, all_signals, sim_item_signals,
                            seqs_with_signal_path: dict):
    all_signal_ids = [signal.id for signal in all_signals]
    signal_matrix = annotated_sequences.get_signal_matrix()

    for signal in sim_item_signals:
        if max_counts[signal.id] > 0:
            if isinstance(signal, SignalPair):
                selection = signal_matrix[:,
                            [all_signal_ids.index(sid) for sid in signal.id.split(Constants.SIGNAL_DELIMITER)]].astype(
                    bool).all(
                    axis=1)
            else:
                selection = signal_matrix[:, all_signal_ids.index(signal.id)].astype(bool)
                selection = np.logical_and(selection, signal_matrix.sum(axis=1) == 1)
            data_to_write = annotated_sequences[selection][:max_counts[signal.id]]
            if len(data_to_write) > 0:
                write_bnp_data(data=data_to_write, path=seqs_with_signal_path[signal.id])
            max_counts[signal.id] -= len(data_to_write)

    return max_counts


def get_signal_sequences(bnp_data_class, used_seq_count: dict, sim_item: SimConfigItem,
                         sequence_paths: Dict[str, Path]):
    sequences = None
    for signal in sim_item.signals:

        skip_rows = used_seq_count[signal.id]
        n_rows = round(sim_item.receptors_in_repertoire_count * sim_item.signal_proportions[signal])

        data = get_bnp_data(sequence_paths[signal.id], bnp_data_class)
        if data is not None:
            sequences_sig = data[skip_rows:skip_rows + n_rows]

            used_seq_count[signal.id] += n_rows

            sequences_sig = assign_duplicate_counts(sequences_sig, signal.clonal_frequency)

            if sequences is None:
                sequences = sequences_sig
            else:
                sequences = merge_dataclass_objects([sequences, sequences_sig])

    return sequences, used_seq_count


def assign_duplicate_counts(sequences, clonal_frequency_params: dict):
    if clonal_frequency_params:
        return sequences.add_fields(
            fields={'duplicate_count': zipf.rvs(**{**clonal_frequency_params, **{'size': len(sequences)}})},
            field_type_map={'duplicate_count': int})
    else:
        return sequences


def get_no_signal_sequences(sequences, used_seq_count: dict, seqs_no_signal_count: int, bnp_data_class,
                            sequence_paths: Dict[str, Path],
                            sim_item: SimConfigItem):
    if sequence_paths['no_signal'].is_file() and seqs_no_signal_count > 0:
        skip_rows = used_seq_count['no_signal']
        used_seq_count['no_signal'] = used_seq_count['no_signal'] + seqs_no_signal_count
        sequences_no_sig = get_bnp_data(sequence_paths['no_signal'], bnp_data_class)[
                           skip_rows:skip_rows + seqs_no_signal_count]
        sequences_no_sig = assign_duplicate_counts(sequences_no_sig, sim_item.default_clonal_frequency)

        if sequences is None:
            sequences = sequences_no_sig
        else:
            sequences = merge_dataclass_objects([sequences, sequences_no_sig])

        return sequences, used_seq_count
    else:
        return sequences, used_seq_count


def needs_seqs_with_signal(sequence_per_signal_count: dict) -> bool:
    return (sum(sequence_per_signal_count.values()) - sequence_per_signal_count['no_signal']) > 0


def filter_sequences_by_length(sequences, sim_item: SimConfigItem, sequence_type):
    sim_item.sequence_len_limits = {} if sim_item.sequence_len_limits is None else sim_item.sequence_len_limits
    region_type = sim_item.generative_model.region_type

    sim_item.sequence_len_limits['min'] = get_min_seq_length(sim_item, sequence_type, region_type)
    sim_item.sequence_len_limits['max'] = get_max_seq_length(sim_item, sequence_type, region_type)
    logging.info(f"Simulation item {sim_item.name}: setting min and max sequence length to "
                 f"{sim_item.sequence_len_limits['min']} and {sim_item.sequence_len_limits['max']} since IMGT "
                 f"numbering will be used downstream for signal annotation or implanting.")

    sequences = sequences[getattr(sequences, sequence_type.value).lengths <= sim_item.sequence_len_limits['max']]
    sequences = sequences[getattr(sequences, sequence_type.value).lengths >= sim_item.sequence_len_limits['min']]

    assert np.all(getattr(sequences, sequence_type.value).lengths <= sim_item.sequence_len_limits['max']), \
        f'An error occurred while filtering sequences by length: some sequences are longer than {sim_item.sequence_len_limits["max"]}'
    assert np.all(getattr(sequences, sequence_type.value).lengths >= sim_item.sequence_len_limits['min']), \
        f'An error occurred while filtering sequences by length: some sequences are shorter than {sim_item.sequence_len_limits["min"]}'

    return sequences


def get_min_seq_length(sim_item: SimConfigItem, sequence_type: SequenceType, region_type: RegionType) -> int:
    conversion_constant = 1 if sequence_type == SequenceType.AMINO_ACID else 3
    if region_type == RegionType.IMGT_JUNCTION:
        return max((PositionHelper.MIN_CDR3_LEN + 2) * conversion_constant, sim_item.sequence_len_limits['min'])
    elif region_type == RegionType.IMGT_CDR3:
        return max(PositionHelper.MIN_CDR3_LEN * conversion_constant, sim_item.sequence_len_limits['min'])
    else:
        raise RuntimeError(f"Unsupported region type for IMGT numbering encountered during simulation: {region_type}.")


def get_max_seq_length(sim_item: SimConfigItem, sequence_type: SequenceType, region_type: RegionType) -> int:
    conversion_constant = 1 if sequence_type == SequenceType.AMINO_ACID else 3

    if region_type == RegionType.IMGT_JUNCTION:
        max_allowed = (PositionHelper.MAX_CDR3_LEN + 2) * conversion_constant
    elif region_type == RegionType.IMGT_CDR3:
        max_allowed = PositionHelper.MAX_CDR3_LEN * conversion_constant
    else:
        raise RuntimeError(f"Unsupported region type for IMGT numbering encountered during simulation: {region_type}.")

    if ('max' not in sim_item.sequence_len_limits or sim_item.sequence_len_limits['max'] == -1 or
            sim_item.sequence_len_limits['max'] > max_allowed):
        return max_allowed
    else:
        return sim_item.sequence_len_limits['max']



import torch


class PyTorchLogisticRegression(torch.nn.Module):

    def __init__(self, in_features: int, zero_abundance_weight_init: bool):
        super().__init__()

        self.linear = torch.nn.Linear(in_features, 1, bias=True)
        with torch.no_grad():
            self.linear.bias.zero_()
            self.linear.weight.normal_(mean=0, std=1/in_features)
            if zero_abundance_weight_init:
                self.linear.weight[:, -1].fill_(0.0)

    def forward(self, x):
        return self.linear(x).squeeze()



import logging
import shutil
from pathlib import Path

import yaml

from immuneML.IO.dataset_export.AIRRExporter import AIRRExporter
from immuneML.api.galaxy.GalaxyTool import GalaxyTool
from immuneML.api.galaxy.Util import Util
from immuneML.app.ImmuneMLApp import ImmuneMLApp
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.workflows.instructions.dataset_generation.DatasetExportInstruction import DatasetExportInstruction


class DataSimulationTool(GalaxyTool):

    def __init__(self, specification_path: Path, result_path: Path, **kwargs):
        Util.check_parameters(specification_path, result_path, kwargs, DataSimulationTool.__name__)
        super().__init__(specification_path, result_path, **kwargs)
        self.expected_instruction = DatasetExportInstruction.__name__[:-11]
        self.instruction_name = None
        self.dataset_name = None

    def _run(self):
        self.prepare_specs()
        # Util.run_tool(self.yaml_path, self.result_path)

        state = ImmuneMLApp(self.yaml_path, self.result_path).run()[0]
        dataset = state.datasets[0]

        Util.export_galaxy_dataset(dataset, self.result_path)

        logging.info(f"{DataSimulationTool.__name__}: immuneML has finished and the dataset was created.")

    def prepare_specs(self):
        with self.yaml_path.open("r") as file:
            specs = yaml.safe_load(file)

        self.instruction_name = Util.check_instruction_type(specs, DataSimulationTool.__name__, self.expected_instruction)

        ParameterValidator.assert_keys_present(specs["definitions"], ["datasets"], DataSimulationTool.__name__, "definitions/datasets")
        ParameterValidator.assert_type_and_value(specs['definitions']['datasets'], dict, DataSimulationTool.__name__, "definitions/datasets")

        self.dataset_name = "dataset"
        Util.update_dataset_key(specs, DataSimulationTool.__name__, self.dataset_name)

        Util.check_paths(specs, DataSimulationTool.__name__)
        Util.update_result_paths(specs, self.result_path, self.yaml_path)



import numpy as np
import pandas as pd
from sklearn.metrics import DistanceMetric


class FurthestNeighborClassifier:
    """
    Furthest Neighbor Classifier for clustering tasks. It predicts the label (cluster) of an example as the label
    corresponding to the minimal maximum distance across all labels in training data. The metric used for distance
    computation can be any metric supported by sklearn.metrics.DistanceMetric, or precomputed (e.g., when using
    TCRdistEncoder).

    """
    def __init__(self, metric: str = 'precomputed', **kwargs):
        super().__init__(**kwargs)
        self.metric = metric

        self.X_train = None
        self.y_train = None
        self.classes_ = None

    def fit(self, X, y):
        self.y_train = y
        self.classes_ = np.unique(y)
        self.X_train = X

        if self.metric == 'precomputed':
            assert X.shape[0] == X.shape[1], \
                (f"{FurthestNeighborClassifier.__name__}: distance matrix must be square for precomputed metric, "
                 f"got: {X.shape}.")

        return self

    def predict(self, X):
        if self.metric != 'precomputed':
            try:
                distance_metric = DistanceMetric.get_metric(self.metric)
            except Exception as e:
                raise ValueError(f"{FurthestNeighborClassifier.__name__}: Metric '{self.metric}' couldn't be "
                                 f"computed. Full error: {e}")
            distances = distance_metric.pairwise(X, self.X_train)
        else:
            distances = X

        max_dist_per_cluster = pd.DataFrame(
            {cluster: distances[:, self.y_train == cluster].max(axis=1) for cluster in self.classes_})
        predictions = max_dist_per_cluster.idxmin(axis=1).values

        return predictions



import logging

from immuneML import Constants
from immuneML.data_model.SequenceSet import ReceptorSequence
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.kmer_frequency.sequence_encoding.SequenceEncodingStrategy import SequenceEncodingStrategy
from immuneML.util.KmerHelper import KmerHelper


class GappedKmerSequenceEncoder(SequenceEncodingStrategy):

    @staticmethod
    def encode_sequence(sequence: ReceptorSequence, params: EncoderParams, encode_locus=False):
        """
        creates all overlapping gapped k-mers from a sequence as features for use in KmerFrequencyEncoder.
        this gap length goes from min_gap to max_gap inclusive, and there is a k-mer of length k_left on the left
        side of the gap and a k-mer of length k_right on the right side of the gap.
        :param sequence: ReceptorSequence
        :param params: EncoderParams (within the "model", the following keys are used: "k_left", "k_right", "max_gap",
                        "min_gap")
        :return: SequenceEncodingResult

        Args:
            encode_locus:
        """
        k_left = params.model.get('k_left')
        k_right = params.model.get('k_right', k_left)
        max_gap = params.model.get('max_gap')
        min_gap = params.model.get('min_gap', 0)
        sequence_type = params.model.get('sequence_type', params.sequence_type)
        length = len(sequence.get_sequence(sequence_type))

        if length < k_left + k_right + max_gap:
            logging.warning('Sequence length is less than k_left + k_right + max_gap. Ignoring sequence')
            return None

        gapped_kmers = KmerHelper.create_gapped_kmers_from_sequence(sequence, k_left=k_left, max_gap=max_gap,
                                                                    min_gap=min_gap, k_right=k_right, sequence_type=sequence_type)

        if encode_locus:
            gapped_kmers = [f"{sequence.locus}{Constants.FEATURE_DELIMITER}{kmer}" for kmer in gapped_kmers]

        return gapped_kmers

    @staticmethod
    def get_feature_names(params: EncoderParams):
        return ["sequence"]



from pathlib import Path
from typing import List

import numpy as np
import pandas as pd

from immuneML.IO.ml_method.UtilIO import UtilIO
from immuneML.caching.CacheHandler import CacheHandler
from immuneML.data_model.EncodedData import EncodedData
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.abundance_encoding.AbundanceEncoderHelper import AbundanceEncoderHelper
from immuneML.pairwise_repertoire_comparison.ComparisonData import ComparisonData
from immuneML.util.EncoderHelper import EncoderHelper


class SequenceAbundanceEncoder(DatasetEncoder):
    """
    This encoder represents the repertoires as vectors where:

    - the first element corresponds to the number of label-associated clonotypes
    - the second element is the total number of unique clonotypes

    To determine what clonotypes (with features defined by comparison_attributes) are label-associated, one-sided Fisher's exact test is used.

    The encoder also writes out files containing the contingency table used for Fisher's exact test,
    the resulting p-values, and the significantly abundant sequences
    (use :py:obj:`~immuneML.reports.encoding_reports.RelevantSequenceExporter.RelevantSequenceExporter` to export these sequences in AIRR format).

    Reference: Emerson, Ryan O. et al.
    ‘Immunosequencing Identifies Signatures of Cytomegalovirus Exposure History and HLA-Mediated Effects on the T Cell Repertoire’.
    Nature Genetics 49, no. 5 (May 2017): 659–65. `doi.org/10.1038/ng.3822 <https://doi.org/10.1038/ng.3822>`_.

    Note: to use this encoder, it is necessary to explicitly define the positive class for the label when defining the label
    in the instruction. With positive class defined, it can then be determined which sequences are indicative of the positive class.
    For full example of using this encoder, see :ref:`Reproduction of the CMV status predictions study`.

    **Dataset type:**

    - RepertoireDatasets

    .. note::

        This encoder is computationally intensive and may require a large amount of memory and time to run. Use
        CompAIRRSequenceAbundance encoder instead for more efficient computation and for the same functionality.


    **Specification arguments:**

    - comparison_attributes (list): The attributes to be considered to group receptors into clonotypes. Only the fields specified in
      comparison_attributes will be considered, all other fields are ignored. Valid comparison value can be any
      repertoire field name (e.g., as specified in the AIRR rearrangement schema).

    - p_value_threshold (float): The p value threshold to be used by the statistical test.

    - sequence_batch_size (int): The number of sequences in a batch when comparing sequences across repertoires, typically 100s of thousands.
      This does not affect the results of the encoding, only the speed. The default value is 1.000.000

    - repertoire_batch_size (int): How many repertoires will be loaded at once. This does not affect the result of the encoding, only the speed.
      This value is a trade-off between the number of repertoires that can fit the RAM at the time and loading time from disk.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            encodings:
                my_sa_encoding:
                    SequenceAbundance:
                        comparison_attributes:
                            - cdr3_aa
                            - v_call
                            - j_call
                        p_value_threshold: 0.05
                        sequence_batch_size: 100000
                        repertoire_batch_size: 32

    """

    RELEVANT_SEQUENCE_ABUNDANCE = "relevant_sequence_abundance"
    TOTAL_SEQUENCE_ABUNDANCE = "total_sequence_abundance"

    def __init__(self, comparison_attributes, p_value_threshold: float, sequence_batch_size: int, repertoire_batch_size: int, name: str = None):
        super().__init__(name=name)
        self.comparison_attributes = comparison_attributes
        self.sequence_batch_size = sequence_batch_size
        self.relevant_sequence_indices = None
        self.context = None
        self.p_value_threshold = p_value_threshold
        self.relevant_indices_path = None
        self.relevant_sequence_path = None
        self.contingency_table_path = None
        self.p_values_path = None
        self.comparison_data = None
        self.repertoire_batch_size = repertoire_batch_size

    @staticmethod
    def build_object(dataset, **params):
        assert isinstance(dataset, RepertoireDataset), \
            "SequenceAbundanceEncoder: this encoding only works on repertoire datasets."
        return SequenceAbundanceEncoder(**params)

    def encode(self, dataset, params: EncoderParams):
        EncoderHelper.check_positive_class_labels(params.label_config, SequenceAbundanceEncoder.__name__)

        self.comparison_data = self._build_comparison_data(dataset, params)
        return self._encode_data(dataset, params)

    def _build_comparison_data(self, dataset: RepertoireDataset, params: EncoderParams):

        current_dataset = EncoderHelper.get_current_dataset(dataset, self.context)
        comparison_data = CacheHandler.memo_by_params(
            EncoderHelper.build_comparison_params(current_dataset, self.comparison_attributes),
            lambda: EncoderHelper.build_comparison_data(current_dataset, params,
                                                        self.comparison_attributes,
                                                        self.sequence_batch_size))

        return comparison_data

    def _encode_data(self, dataset: RepertoireDataset, params: EncoderParams):

        label_name = params.label_config.get_labels_by_name()[0]

        examples = self._calculate_abundance_matrix(dataset, self.comparison_data, params)

        encoded_data = EncodedData(examples, dataset.get_metadata([label_name]) if params.encode_labels else None, dataset.get_repertoire_ids(),
                                   [SequenceAbundanceEncoder.RELEVANT_SEQUENCE_ABUNDANCE, SequenceAbundanceEncoder.TOTAL_SEQUENCE_ABUNDANCE],
                                   encoding=SequenceAbundanceEncoder.__name__, info={'relevant_sequence_path': self.relevant_sequence_path,
                                                                                     "contingency_table_path": self.contingency_table_path,
                                                                                     "p_values_path": self.p_values_path})

        encoded_dataset = dataset.clone()
        encoded_dataset.encoded_data = encoded_data

        return encoded_dataset

    def _calculate_abundance_matrix(self, dataset: RepertoireDataset, comparison_data: ComparisonData, params: EncoderParams):
        comparison_data.set_iteration_repertoire_ids(dataset.get_repertoire_ids())
        is_positive_class = AbundanceEncoderHelper.check_is_positive_class(dataset, dataset.get_repertoire_ids(), params.label_config)

        relevant_sequence_indices, file_paths = AbundanceEncoderHelper.get_relevant_sequence_indices(comparison_data, is_positive_class,
                                                                                                     self.p_value_threshold,
                                                                                                     self.relevant_indices_path, params,
                                                                                                     cache_params=(dataset.get_repertoire_ids(),
                                                                                                                   self.comparison_attributes))

        self._write_relevant_sequences_csv(comparison_data, relevant_sequence_indices, params.result_path)
        self._set_file_paths(file_paths)

        abundance_matrix = self._build_abundance_matrix(comparison_data, dataset.get_repertoire_ids(), relevant_sequence_indices)

        return abundance_matrix

    def _write_relevant_sequences_csv(self, comparison_data, relevant_sequence_indices, result_path):
        if self.relevant_sequence_path is None:
            self.relevant_sequence_path = result_path / 'relevant_sequences.csv'

        all_sequences = comparison_data.get_item_names()
        relevant_sequences = all_sequences[relevant_sequence_indices]
        if relevant_sequences is not None and sum(relevant_sequence_indices) > 0:
            df = pd.DataFrame({attr: relevant_sequences[:, i] for i, attr in enumerate(self.comparison_attributes)})
        else:
            df = pd.DataFrame(columns=self.comparison_attributes)
        sequence_csv_path = result_path / 'relevant_sequences.csv'
        df.to_csv(sequence_csv_path, sep=',', index=False)

    def _set_file_paths(self, file_paths):
        self.relevant_indices_path = file_paths["relevant_indices_path"]
        self.contingency_table_path = file_paths["contingency_table_path"] if "contingency_table_path" in file_paths else None
        self.p_values_path = file_paths["p_values_path"] if "p_values_path" in file_paths else None

    def _build_abundance_matrix(self, comparison_data, repertoire_ids, relevant_sequence_indices):
        abundance_matrix = np.zeros((len(repertoire_ids), 2))

        for index in range(0, len(repertoire_ids) + self.repertoire_batch_size, self.repertoire_batch_size):
            ind_start, ind_end = index, min(index + self.repertoire_batch_size, len(repertoire_ids))
            repertoire_vectors = comparison_data.get_repertoire_vectors(repertoire_ids[ind_start:ind_end])

            for rep_index in range(ind_start, ind_end):
                repertoire_vector = repertoire_vectors[repertoire_ids[rep_index]]
                relevant_sequence_abundance = np.sum(
                    repertoire_vector[np.logical_and(relevant_sequence_indices, repertoire_vector)])
                total_sequence_abundance = np.sum(repertoire_vector)
                abundance_matrix[rep_index] = [relevant_sequence_abundance, total_sequence_abundance]

        return abundance_matrix

    def set_context(self, context: dict):
        self.context = context
        return self

    def get_additional_files(self) -> List[Path]:
        return [self.relevant_indices_path]

    @staticmethod
    def load_encoder(encoder_file: Path):
        encoder = DatasetEncoder.load_encoder(encoder_file)
        encoder.relevant_indices_path = DatasetEncoder.load_attribute(encoder, encoder_file, "relevant_indices_path")
        encoder.comparison_data = UtilIO.import_comparison_data(encoder_file.parent)
        return encoder



from pathlib import Path

import yaml

from immuneML.api.galaxy.GalaxyTool import GalaxyTool
from immuneML.api.galaxy.Util import Util
from immuneML.app.ImmuneMLApp import ImmuneMLApp
from immuneML.util.Logger import print_log
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.exploratory_analysis.ExploratoryAnalysisInstruction import \
    ExploratoryAnalysisInstruction


class DatasetGenerationOverviewTool(GalaxyTool):
    """
    DatasetGenerationOverviewTool is an alternative to running ImmuneMLApp directly.
    This tool is meant to be used as an endpoint for Galaxy tool that will create a Galaxy collection out of a dataset in immuneML format.

    This tool accepts a path to a YAML specification which uses a single dataset, and runs the ExploratoryAnalysisInstruction with optional reports.
    The created dataset will be located in the supplied output directory, under the 'galaxy_dataset' folder.
    The main dataset file will have the name of the dataset given in the specification and has an extension .yaml.
    """

    def __init__(self, specification_path: Path, result_path: Path, **kwargs):
        Util.check_parameters(specification_path, result_path, kwargs, "Dataset generation tool")
        super().__init__(specification_path, result_path, **kwargs)

    def _run(self):
        PathBuilder.build(self.result_path)
        self._update_specs()
        state = ImmuneMLApp(self.yaml_path, self.result_path).run()[0]
        dataset = list(state.exploratory_analysis_units.values())[0].dataset

        Util.export_galaxy_dataset(dataset, self.result_path)

        print_log(f"Exported dataset.")

    def _update_specs(self):
        with self.yaml_path.open('r') as file:
            specs = yaml.safe_load(file)

        ParameterValidator.assert_keys_present(specs.keys(), ["definitions", "instructions"], DatasetGenerationOverviewTool.__name__, "YAML specification")
        ParameterValidator.assert_all_in_valid_list(specs.keys(), ["definitions", "instructions", "output"], DatasetGenerationOverviewTool.__name__, "YAML specification")

        self._check_dataset(specs)
        self._check_instruction(specs)

        Util.update_dataset_key(specs, DatasetGenerationOverviewTool.__name__)
        Util.check_paths(specs, DatasetGenerationOverviewTool.__name__)
        Util.update_result_paths(specs, self.result_path, self.yaml_path)

    def _check_dataset(self, specs):
        ParameterValidator.assert_keys_present(specs["definitions"].keys(), ['datasets'], DatasetGenerationOverviewTool.__name__, 'definitions')
        assert len(specs['definitions']['datasets'].keys()) == 1, \
            f"{DatasetGenerationOverviewTool.__name__}: only one dataset can be defined with this Galaxy tool, got these " \
            f"instead: {list(specs['definitions']['datasets'].keys())}."

    def _check_instruction(self, specs):
        assert len(specs['instructions'].keys()) == 1, \
            f"{DatasetGenerationOverviewTool.__name__}: only one instruction of type ExploratoryAnalysis can be defined with this Galaxy tool, got these " \
            f"instructions instead: {list(specs['instructions'].keys())}."

        instruction_name = Util.check_instruction_type(specs, DatasetGenerationOverviewTool.__name__, ExploratoryAnalysisInstruction.__name__[:-11])

        dataset_name = None
        for analysis_key, analysis_specs in specs['instructions'][instruction_name]["analyses"].items():
            ParameterValidator.assert_keys_present(analysis_specs.keys(), ["dataset", "report"],
                                                   DatasetGenerationOverviewTool.__name__, f"{instruction_name}/analyses/{analysis_key}")

            if dataset_name is None:
                dataset_name = analysis_specs["dataset"]
            else:
                assert analysis_specs["dataset"] == dataset_name, f"{DatasetGenerationOverviewTool.__name__}: expected only one dataset name. Found: {dataset_name} and {analysis_specs['dataset']}"



import os
import shutil
from unittest import TestCase

from immuneML.caching.CacheType import CacheType
from immuneML.environment.Constants import Constants
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.preprocessing.filters.SillyFilter import SillyFilter
from immuneML.simulation.dataset_generation.RandomDatasetGenerator import RandomDatasetGenerator


class TestSillyFilter(TestCase):

    def setUp(self) -> None:
        os.environ[Constants.CACHE_TYPE] = CacheType.TEST.name

    def _get_mock_repertoire_dataset(self, path):
        # Create a mock RepertoireDataset with 10 repertoires, each containing 50 sequences of length 15,
        dataset = RandomDatasetGenerator.generate_repertoire_dataset(repertoire_count=10,
                                                                     sequence_count_probabilities={50: 1},
                                                                     sequence_length_probabilities={15: 1},
                                                                     labels={},
                                                                     path=path)

        return dataset

    def test_process_dataset(self):
        tmp_path = EnvironmentSettings.tmp_test_path / "silly_filter/"

        dataset = self._get_mock_repertoire_dataset(tmp_path / "original_dataset")

        params = {"fraction_to_keep": 0.8}
        filter = SillyFilter.build_object(**params)

        processed_dataset = filter.process_dataset(dataset, tmp_path / "filtered_dataset")

        # 10 original repertoires, keep 80%
        assert len(processed_dataset.repertoires) == 8

        shutil.rmtree(tmp_path)



import hashlib
import logging
import os
import pickle
from pathlib import Path

import dill

from immuneML.caching.CacheObjectType import CacheObjectType
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.util.PathBuilder import PathBuilder


class CacheHandler:

    @staticmethod
    def get_file_path(cache_type=None):
        file_path = EnvironmentSettings.get_cache_path(cache_type) / "files"
        PathBuilder.build(file_path)
        return file_path

    @staticmethod
    def get(params: tuple, object_type, cache_type=None):
        h = CacheHandler.hash(params)
        return CacheHandler.get_by_key(h, object_type, cache_type)

    @staticmethod
    def get_by_key(cache_key: str, object_type, cache_type=None):
        filename = CacheHandler._build_filename(cache_key, object_type, cache_type)
        obj = None
        if filename.is_file():
            with filename.open("rb") as file:
                obj = dill.load(file)
        return obj

    @staticmethod
    def _build_filename(cache_key: str, object_type: CacheObjectType, cache_type=None) -> Path:
        path = EnvironmentSettings.get_cache_path(cache_type) / object_type.name.lower()
        PathBuilder.build(path)
        return path / f"{cache_key}.pickle"

    @staticmethod
    def exists(params: tuple, object_type: CacheObjectType = CacheObjectType.OTHER, cache_type=None):
        return CacheHandler.get_by_key(CacheHandler.generate_cache_key(params), object_type, cache_type) is not None

    @staticmethod
    def add(params: tuple, caching_object, object_type: CacheObjectType = CacheObjectType.OTHER, cache_type=None):
        PathBuilder.build(EnvironmentSettings.get_cache_path(cache_type))
        h = CacheHandler.generate_cache_key(params)
        filename = CacheHandler._build_filename(cache_key=h, object_type=object_type, cache_type=cache_type)
        with filename.open("wb") as file:
            dill.dump(caching_object, file, protocol=pickle.HIGHEST_PROTOCOL)

    @staticmethod
    def add_by_key(cache_key: str, caching_object, object_type: CacheObjectType = CacheObjectType.OTHER, cache_type=None):
        PathBuilder.build(EnvironmentSettings.get_cache_path(cache_type))
        filename = CacheHandler._build_filename(cache_key=cache_key, object_type=object_type, cache_type=cache_type)
        try:
            with filename.open("wb") as file:
                dill.dump(caching_object, file, protocol=pickle.HIGHEST_PROTOCOL)
        except AttributeError:
            os.remove(filename)
            logging.warning(f"CacheHandler: could not cache object of class {type(caching_object).__name__} with key {cache_key}. "
                            f"Object: {caching_object}\n"
                            f"Next time this object is needed, it will be recomputed which will take more time but should not influence results.")

    @staticmethod
    def generate_cache_key(params: tuple):
        return hashlib.sha256(str(params).encode('utf-8')).hexdigest()

    @staticmethod
    def memo(cache_key: str, fn, object_type: CacheObjectType = CacheObjectType.OTHER, cache_type=None):
        result = CacheHandler.get_by_key(cache_key, object_type, cache_type)
        if result is None:
            result = fn()
            CacheHandler.add_by_key(cache_key, result, object_type, cache_type)
        return result

    @staticmethod
    def memo_by_params(params: tuple, fn, object_type: CacheObjectType = CacheObjectType.OTHER, cache_type=None):
        cache_key = CacheHandler.generate_cache_key(params)
        return CacheHandler.memo(cache_key, fn, object_type, cache_type)

    @staticmethod
    def hash(params: tuple) -> str:
        return hashlib.sha256(str(params).encode('utf-8')).hexdigest()



from pathlib import Path

import pandas as pd

from immuneML.data_model import bnp_util
from immuneML.data_model.SequenceSet import Repertoire
from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.environment.Constants import Constants
from immuneML.preprocessing.Preprocessor import Preprocessor
from immuneML.util.PathBuilder import PathBuilder


class SubjectRepertoireCollector(Preprocessor):
    """
    Merges all the Repertoires in a RepertoireDataset that have the same 'subject_id' specified in the metadata. The result
    is a RepertoireDataset with one Repertoire per subject. This preprocessing cannot be used in combination with :ref:`TrainMLModel`
    instruction because it can change the number of examples. To combine the repertoires in this way, use this preprocessing
    with :ref:`DatasetExport` instruction.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        preprocessing_sequences:
            my_preprocessing:
                - my_filter: SubjectRepertoireCollector

    """

    def __init__(self, result_path: Path = None):
        super().__init__(result_path)

    def process_dataset(self, dataset: RepertoireDataset, result_path: Path, number_of_processes=1):
        self.result_path = PathBuilder.build(result_path if result_path is not None else self.result_path)
        self.check_dataset_type(dataset, [RepertoireDataset], "SubjectRepertoireCollector")

        processed_dataset = self._merge_repertoires(dataset)

        return processed_dataset

    def _merge_repertoires(self, dataset: RepertoireDataset):
        rep_map = {}
        repertoires, indices_to_keep = [], []
        processed_dataset = dataset.clone()

        for index, repertoire in enumerate(processed_dataset.get_data()):

            if repertoire.metadata["subject_id"] in rep_map.keys():
                rep_map[repertoire.metadata['subject_id']].append(repertoire)
            else:
                rep_map[repertoire.metadata['subject_id']] = [repertoire]

        for key in rep_map.keys():
            repertoires.append(self._store_repertoire(rep_map[key]))

        processed_dataset.repertoires = repertoires
        processed_dataset.metadata_file = self._build_new_metadata(dataset, indices_to_keep)

        return processed_dataset

    def _build_new_metadata(self, dataset, indices_to_keep):
        if dataset.metadata_file:
            df = pd.read_csv(dataset.metadata_file, index_col=0, comment=Constants.COMMENT_SIGN).iloc[indices_to_keep,
                 :]
            path = Path(self.result_path / f"{dataset.metadata_file.stem}_collected_repertoires.csv")
            df.to_csv(path)
        else:
            path = None
        return path

    def _store_repertoire(self, repertoires: list):
        metadata = {k: v for d in [r.metadata for r in repertoires] for k, v in d.items()}
        new_repertoire = Repertoire.build_from_dc_object(self.result_path, metadata,
                                                         data=bnp_util.merge_dataclass_objects(
                                                             [r.data for r in repertoires]))
        return new_repertoire

    def keeps_example_count(self) -> bool:
        return False



from pathlib import Path
from typing import List

from immuneML.data_model.datasets.Dataset import Dataset
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.reports.Report import Report
from immuneML.reports.ReportResult import ReportResult
from immuneML.util.Logger import print_log
from immuneML.util.PathBuilder import PathBuilder
from immuneML.workflows.instructions.Instruction import Instruction
from immuneML.workflows.instructions.exploratory_analysis.ExploratoryAnalysisState import ExploratoryAnalysisState
from immuneML.workflows.instructions.exploratory_analysis.ExploratoryAnalysisUnit import ExploratoryAnalysisUnit
from immuneML.workflows.steps.DataEncoder import DataEncoder
from immuneML.workflows.steps.DataEncoderParams import DataEncoderParams
from immuneML.workflows.steps.DataWeighter import DataWeighter
from immuneML.workflows.steps.DataWeighterParams import DataWeighterParams


class ExploratoryAnalysisInstruction(Instruction):
    """
    Allows exploratory analysis of different datasets using encodings and reports.

    Analysis is defined by a dictionary of ExploratoryAnalysisUnit objects that encapsulate a dataset, an encoding [optional]
    and a report to be executed on the [encoded] dataset. Each analysis specified under `analyses` is completely independent from all
    others.

    .. note::

        The "report" parameter has been updated to support multiple "reports" per analysis unit. For backward
        compatibility, the "report" key is still accepted, but it will be ignored if "reports" is present.
        "report" option will be removed in the next major version.

    **Specification arguments:**

    - analyses (dict): a dictionary of analyses to perform. The keys are the names of different analyses, and the values for each
      of the analyses are:

      - dataset: dataset on which to perform the exploratory analysis

      - preprocessing_sequence: which preprocessings to use on the dataset, this item is optional and does not have to be specified.

      - example_weighting: which example weighting strategy to use before encoding the data, this item is optional and does not have to be specified.

      - encoding: how to encode the dataset before running the report, this item is optional and does not have to be specified.

      - labels: if encoding is specified, the relevant labels should be specified here.

      - dim_reduction: which dimensionality reduction to apply;

      - reports: which reports to run on the dataset. Reports specified here may be of the category :ref:`**Data reports**`
        or :ref:`**Encoding reports**`, depending on whether 'encoding' was specified.

    - number_of_processes: (int): how many processes should be created at once to speed up the analysis. For personal
      machines, 4 or 8 is usually a good choice.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        instructions:
            my_expl_analysis_instruction: # user-defined instruction name
                type: ExploratoryAnalysis # which instruction to execute
                analyses: # analyses to perform
                    my_first_analysis: # user-defined name of the analysis
                        dataset: d1 # dataset to use in the first analysis
                        preprocessing_sequence: p1 # preprocessing sequence to use in the first analysis
                        reports: [r1] # which reports to generate using the dataset d1
                    my_second_analysis: # user-defined name of another analysis
                        dataset: d1 # dataset to use in the second analysis - can be the same or different from other analyses
                        encoding: e1 # encoding to apply on the specified dataset (d1)
                        reports: [r2] # which reports to generate in the second analysis
                        labels: # labels present in the dataset d1 which will be included in the encoded data on which report r2 will be run
                            - celiac # name of the first label as present in the column of dataset's metadata file
                            - CMV # name of the second label as present in the column of dataset's metadata file
                    my_third_analysis: # user-defined name of another analysis
                        dataset: d1 # dataset to use in the second analysis - can be the same or different from other analyses
                        encoding: e1 # encoding to apply on the specified dataset (d1)
                        dim_reduction: umap # or None; which dimensionality reduction method to apply to encoded d1
                        reports: [r3] # which report to generate in the third analysis
                number_of_processes: 4 # number of parallel processes to create (could speed up the computation)
    """

    def __init__(self, exploratory_analysis_units: dict, name: str = None):
        assert all(isinstance(unit, ExploratoryAnalysisUnit) for unit in exploratory_analysis_units.values()), \
            ("ExploratoryAnalysisInstruction: not all elements passed to init method are instances of "
             "ExploratoryAnalysisUnit.")
        self.state = ExploratoryAnalysisState(exploratory_analysis_units, name=name)
        self.name = name

    def run(self, result_path: Path):
        name = self.name if self.name is not None else "exploratory_analysis"
        self.state.result_path = result_path / name
        for index, (key, unit) in enumerate(self.state.exploratory_analysis_units.items()):
            print_log(f"Started analysis {key} ({index + 1}/{len(self.state.exploratory_analysis_units)}).",
                      include_datetime=True)
            path = self.state.result_path / f"analysis_{key}"
            PathBuilder.build(path)
            report_results = self.run_unit(unit, path)
            unit.report_results = report_results
            print_log(f"Finished analysis {key} ({index + 1}/{len(self.state.exploratory_analysis_units)}).\n",
                      include_datetime=True)
        return self.state

    def run_unit(self, unit: ExploratoryAnalysisUnit, result_path: Path) -> List[ReportResult]:
        unit.dataset = self.preprocess_dataset(unit, result_path / "preprocessed_dataset")
        unit.dataset = self.weight_examples(unit, result_path / "weighted_dataset")
        unit.dataset = self.encode(unit, result_path / "encoded_dataset")

        if unit.dim_reduction is not None:
            self._run_dimensionality_reduction(unit)

        report_results = []

        if unit.reports is not None:
            for report in unit.reports:
                if report is not None:
                    report_result = self.run_report(unit, result_path, report)
                    report_results.append(report_result)
        else:
            report_results = None

        return report_results

    def _run_dimensionality_reduction(self, unit: ExploratoryAnalysisUnit):
        result = unit.dim_reduction.fit_transform(unit.dataset)
        unit.dataset.encoded_data.dimensionality_reduced_data = result
        unit.dataset.encoded_data.dim_names = unit.dim_reduction.get_dimension_names()

    def preprocess_dataset(self, unit: ExploratoryAnalysisUnit, result_path: Path) -> Dataset:
        if unit.preprocessing_sequence is not None and len(unit.preprocessing_sequence) > 0:
            dataset = unit.dataset
            for preprocessing in unit.preprocessing_sequence:
                dataset = preprocessing.process_dataset(dataset, result_path,
                                                        number_of_processes=unit.number_of_processes)
                print_log(
                    f"Preprocessed {dataset.__class__.__name__.split('Dataset')[0].lower()} dataset "
                    f"{dataset.name} with {preprocessing.__class__.__name__}:\n"
                    f"- Example count: {dataset.get_example_count()}\n"
                    f"- Labels: {dataset.get_label_names()}", True)
        else:
            dataset = unit.dataset
        return dataset

    def weight_examples(self, unit: ExploratoryAnalysisUnit, result_path: Path):
        if unit.example_weighting is not None:
            from immuneML.example_weighting.ExampleWeightingParams import ExampleWeightingParams
            weighted_dataset = DataWeighter.run(
                DataWeighterParams(dataset=unit.dataset, weighting_strategy=unit.example_weighting,
                                   weighting_params=ExampleWeightingParams(result_path=result_path,
                                                                           pool_size=unit.number_of_processes,
                                                                           learn_model=True),
                                   ))
        else:
            weighted_dataset = unit.dataset

        return weighted_dataset

    def encode(self, unit: ExploratoryAnalysisUnit, result_path: Path) -> Dataset:
        if unit.encoder is not None:
            encoded_dataset = DataEncoder.run(DataEncoderParams(dataset=unit.dataset, encoder=unit.encoder,
                                                                encoder_params=EncoderParams(result_path=result_path,
                                                                                             label_config=unit.label_config,
                                                                                             pool_size=unit.number_of_processes,
                                                                                             learn_model=True,
                                                                                             encode_labels=unit.label_config is not None),
                                                                ))
        else:
            encoded_dataset = unit.dataset
        return encoded_dataset

    def run_report(self, unit: ExploratoryAnalysisUnit, result_path: Path, report: Report) -> ReportResult:
        if (result_path / f"report_{report.name}").exists():
            report_result_path = result_path / f"report_{report.name}_{len(list(result_path.glob(f'report_{report.name}*')))}"
        else:
            report_result_path = result_path / f"report_{report.name}"
        report.result_path = report_result_path
        report.number_of_processes = unit.number_of_processes

        report.dataset = unit.dataset
        return report.generate_report()



from pathlib import Path
from typing import List, Tuple

import numpy as np
import plotly.express as px

from immuneML.data_model.datasets.RepertoireDataset import RepertoireDataset
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.encoding_reports.FeatureReport import FeatureReport


class FeatureValueBarplot(FeatureReport):
    """
    Encoding a dataset results in a numeric matrix, where the rows are examples (e.g., sequences, receptors, repertoires)
    and the columns are features. For example, when :ref:`KmerFrequency` encoder is used, the features are the
    k-mers (AAA, AAC, etc..) and the feature values are the frequencies per k-mer.

    This report plots the mean feature values per feature.
    A bar plot is created where the average feature value across all examples is shown, with optional error bars.
    The bar plots can be separated into different colors or facets using metadata labels
    (for example: plot the average feature values of 'cohort1', 'cohort2' and 'cohort3' in different colors to spot biases).

    See also: :py:obj:`~immuneML.reports.encoding_reports.FeatureDistribution.FeatureDistribution` report to plot
    the distribution of each feature across examples, rather than only showin the mean value in a bar plot.
    Or :py:obj:`~immuneML.reports.encoding_reports.FeatureDistribution.FeatureComparison` report to compare
    features across binary metadata labels (e.g., plot the feature value of 'sick' repertoires on the x axis,
    and 'healthy' repertoires on the y axis.).


    Example output:

    .. image:: ../../_static/images/reports/feature_value_barplot.png
       :alt: Feature value barplot report example
       :width: 750


    **Specification arguments:**

    - color_grouping_label (str): The label that is used to color each bar, at each level of the grouping_label.

    - row_grouping_label (str): The label that is used to group bars into different row facets.

    - column_grouping_label (str): The label that is used to group bars into different column facets.

    - show_error_bar (bool): Whether to show the error bar (standard deviation) for the bars.

    - x_title (str): x-axis label

    - y_title (str): y-axis label

    - plot_top_n (int): plot n of the largest features on average separately (useful when there are too many features
      to plot at the same time). The n features are chosen based on the average feature values across all examples
      without grouping by labels. The plot shows averages per label classes.

    - plot_bottom_n (int): plot n of the smallest features on average separately (useful when there are too many
      features to plot at the same time). The n features are chosen based on the average feature values across all
      examples without grouping by labels. The plot shows averages per label classes.

    - plot_all_features (bool): whether to plot all (might be slow for large number of features)

    - error_function (str): which error function to use for the error bar. Options are 'std' (standard deviation) or
      'sem' (standard error of the mean). Default: std.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_fvb_report:
                    FeatureValueBarplot: # timepoint, disease_status and age_group are metadata labels
                        column_grouping_label: timepoint
                        row_grouping_label: disease_status
                        color_grouping_label: age_group
                        plot_all_features: true
                        plot_top_n: 10
                        plot_bottom_n: 5

    """

    @classmethod
    def build_object(cls, **kwargs):
        return FeatureValueBarplot(**kwargs)

    def __init__(self, dataset: RepertoireDataset = None, result_path: Path = None,
                 color_grouping_label: str = None, row_grouping_label=None, column_grouping_label=None,
                 x_title: str = None, y_title: str = None, show_error_bar=True, name: str = None,
                 plot_all_features: bool = True, error_function: str = None,
                 number_of_processes: int = 1, plot_top_n: int = None, plot_bottom_n: int = None):
        super().__init__(dataset=dataset, result_path=result_path, color_grouping_label=color_grouping_label,
                         row_grouping_label=row_grouping_label, column_grouping_label=column_grouping_label,
                         name=name, number_of_processes=number_of_processes, error_function=error_function)
        self.show_error_bar = show_error_bar
        self.x_title = x_title if x_title is not None else self.x
        self.y_title = y_title if y_title is not None else "value"
        self.result_name = "feature_value_barplot"
        self.name = name
        self.plot_all_features = plot_all_features
        self.plot_top_n = plot_top_n
        self.plot_bottom_n = plot_bottom_n

    def _generate(self):
        result = self._generate_report_result()
        result.info = "A barplot of the feature values in a given encoded data matrix, averaged across examples. Each bar in the barplot represents the mean value of a given feature, and along the x-axis are the different features."
        return result

    def _plot(self, data_long_format) -> Tuple[List[ReportOutput], List[ReportOutput]]:
        groupby_cols_features = [self.x]
        data_groupedby_features = self._get_grouped_data(data_long_format, groupby_cols_features)
        groupby_cols_labels = [self.x, self.color, self.facet_row, self.facet_column]
        data_groupedby_labels = self._get_grouped_data(data_long_format, groupby_cols_labels)

        plotting_data_dict = {'all': data_groupedby_labels} if self.plot_all_features else {}

        error_y = f"value{self.error_function}" if self.show_error_bar else None
        output_figures = []
        output_tables = []

        if self.plot_top_n:
            top_n_features = data_groupedby_features.iloc[
                np.argpartition(data_groupedby_features['valuemean'].values, -self.plot_top_n)[-self.plot_top_n:]][self.x]
            plotting_data_dict[f'top_{self.plot_top_n}'] = data_groupedby_labels.loc[
                data_groupedby_labels[self.x].isin(top_n_features)]

        if self.plot_bottom_n:
            bottom_n_features = data_groupedby_features.iloc[
                np.argpartition(data_groupedby_features['valuemean'].values, self.plot_bottom_n)[:self.plot_bottom_n]][
                self.x]
            plotting_data_dict[f'bottom_{self.plot_bottom_n}'] = data_groupedby_labels.loc[
                data_groupedby_labels[self.x].isin(bottom_n_features)]

        for key, data in plotting_data_dict.items():
            figure = px.bar(data, x=self.x, y="valuemean", color=self.color, barmode="group",
                            facet_row=self.facet_row, facet_col=self.facet_column, error_y=error_y,
                            labels={
                                "valuemean": self.y_title,
                                self.x: self.x_title,
                            }, template='plotly_white',
                            color_discrete_sequence=px.colors.diverging.Tealrose)
            figure.update_layout(xaxis={'categoryorder': 'total descending'})

            file_path = self.result_path / f"{self.result_name}_{key}.html"

            file_path = PlotlyUtil.write_image_to_file(figure, file_path)
            data.to_csv(self.result_path / f"{self.result_name}_{key}.csv", index=False)
            output_tables.append(ReportOutput(path=self.result_path / f"{self.result_name}_{key}.csv", name=f"{self.result_name} {key}"))

            output_figures.append(ReportOutput(path=file_path, name=f"Average feature values ({key})"))

        return output_figures, output_tables



from pathlib import Path

import logging
import plotly.express as px
import pandas as pd
from typing import List

from immuneML.data_model import bnp_util
from immuneML.data_model.datasets.ElementDataset import SequenceDataset
from immuneML.encodings.motif_encoding.PositionalMotifHelper import (
    PositionalMotifHelper,
)
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.environment.SequenceType import SequenceType
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.encodings.motif_encoding.MotifEncoder import MotifEncoder
from immuneML.reports.encoding_reports.EncodingReport import EncodingReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class PositionalMotifFrequencies(EncodingReport):
    """
    This report must be used in combination with the :py:obj:`~immuneML.encodings.motif_encoding.MotifEncoder.MotifEncoder`.
    Plots a stacked bar plot of amino acid occurrence at different indices in any given dataset, along with a plot
    investigating motif continuity which displays a bar plot of the gap sizes between the amino acids in the motifs in
    the given dataset. Note that a distance of 1 means that the amino acids are continuous (next to each other).

    **Specification arguments:**

    - motif_color_map (dict): Optional mapping between motif lengths and specific colors to be used. Example:

        motif_color_map:
            1: #66C5CC
            2: #F6CF71
            3: #F89C74


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            reports:
                my_pos_motif_report:
                    PositionalMotifFrequencies:
                        motif_color_map:

    """

    @classmethod
    def build_object(cls, **kwargs):
        if "motif_color_map" in kwargs:
            ParameterValidator.assert_type_and_value(kwargs["motif_color_map"], dict,
                                                     PositionalMotifFrequencies.__name__, "motif_color_map")
            kwargs["motif_color_map"] = {str(key): value for key, value in kwargs["motif_color_map"].items()}

        return PositionalMotifFrequencies(**kwargs)

    def __init__(self, dataset: SequenceDataset = None, result_path: Path = None,
                 motif_color_map: dict = None, name: str = None, number_of_processes: int = 1):
        super().__init__(dataset=dataset, result_path=result_path, name=name, number_of_processes=number_of_processes)
        self.motif_color_map = motif_color_map
        self.region_type = None
        self.sequence_type = None

    def get_sequence_length(self):
        if self.sequence_type is None or self.region_type is None:
            self.sequence_type = self.dataset.encoded_data.info['sequence_type']
            self.region_type = self.dataset.encoded_data.info['region_type']

        my_sequences = getattr(self.dataset.data, bnp_util.get_sequence_field_name(self.region_type, self.sequence_type))
        return my_sequences.lengths[0]

    def _generate(self):
        PathBuilder.build(self.result_path)

        motifs = self.dataset.encoded_data.feature_names
        positional_aa_counts_df = self._get_positional_aa_counts(motifs)
        max_gap_size_df = self._get_max_gap_sizes(motifs)
        total_gap_size_df = self._get_total_gap_sizes(motifs)

        positional_aa_counts_table = self._write_output_table(positional_aa_counts_df,
                                                              file_path=self.result_path / f"positional_aa_counts.csv",
                                                              name="Frequencies of amino acids found in the high-precision high-recall motifs")

        max_gap_size_table = self._write_output_table(max_gap_size_df,
                                                      file_path=self.result_path / f"max_gap_size_table.csv",
                                                      name="Maximum gap sizes within motifs")

        total_gap_size_table = self._write_output_table(total_gap_size_df,
                                                        file_path=self.result_path / f"total_gap_size_table.csv",
                                                        name="Total (summed) gap sizes within motifs")

        output_figure1 = self._safe_plot(positional_aa_counts_df=positional_aa_counts_df,
                                         plot_callable="_plot_positional_aa_counts")
        output_figure2 = self._safe_plot(gap_size_df=max_gap_size_df, x="max_gap_size", plot_callable="_plot_gap_sizes")
        output_figure3 = self._safe_plot(gap_size_df=total_gap_size_df, x="total_gap_size",
                                         plot_callable="_plot_gap_sizes")

        return ReportResult(
            name=self.name,
            output_figures=[fig for fig in [output_figure1, output_figure2, output_figure3] if fig is not None],
            output_tables=[max_gap_size_table, total_gap_size_table, positional_aa_counts_table],
        )

    def _get_total_gap_sizes(self, motifs):
        data = {"motif_size": [],
                "total_gap_size": [],
                "occurrence": []}

        gap_size_count = {}

        for motif in motifs:
            motif_indices, amino_acids = PositionalMotifHelper.string_to_motif(motif, "&", "-")
            total_gap_size = (max(motif_indices) - min(motif_indices)) - len(motif_indices) + 1
            motif_size = len(motif_indices)

            if motif_size not in gap_size_count:
                gap_size_count[motif_size] = {total_gap_size: 1}
            else:
                if total_gap_size not in gap_size_count[motif_size]:
                    gap_size_count[motif_size][total_gap_size] = 1
                else:
                    gap_size_count[motif_size][total_gap_size] += 1

        for motif_size, counts in gap_size_count.items():
            for total_gap_size, occurrence in counts.items():
                data["motif_size"].append(motif_size)
                data["total_gap_size"].append(total_gap_size)
                data["occurrence"].append(occurrence)

        return pd.DataFrame(data)

    def _get_max_gap_sizes(self, motifs):
        gap_size_count = {
            motif_size: {
                gap_size: 0 for gap_size in range(0, self.get_sequence_length() - 1)
            }
            for motif_size in range(self.get_sequence_length())
        }

        for motif in motifs:
            indices, amino_acids = PositionalMotifHelper.string_to_motif(motif, "&", "-")
            motif_size = len(indices)

            if motif_size > 1:
                gap_size = max([indices[i + 1] - indices[i] - 1 for i in range(motif_size - 1)])
                gap_size_count[motif_size][gap_size] += 1

        motif_sizes = list()
        max_gap_sizes = list()
        occurrence = list()
        for motif_size in gap_size_count:
            if sum(gap_size_count[motif_size].values()) != 0:
                for gap_size in gap_size_count[motif_size]:
                    motif_sizes.append(str(motif_size))
                    max_gap_sizes.append(gap_size)
                    occurrence.append(gap_size_count[motif_size][gap_size])

        gap_size_df = pd.DataFrame()
        gap_size_df["motif_size"] = motif_sizes
        gap_size_df["max_gap_size"] = max_gap_sizes
        gap_size_df["occurrence"] = occurrence

        return gap_size_df

    def _get_positional_aa_counts(self, motifs):
        positional_aa_counts = {
            aa: [0 for i in range(self.get_sequence_length())]
            for aa in EnvironmentSettings.get_sequence_alphabet(SequenceType.AMINO_ACID)
        }

        for motif in motifs:
            indices, amino_acids = PositionalMotifHelper.string_to_motif(
                motif, "&", "-"
            )

            for index, amino_acid in zip(indices, amino_acids):
                positional_aa_counts[amino_acid][index] += 1

        positional_aa_counts_df = pd.DataFrame(positional_aa_counts)
        positional_aa_counts_df = positional_aa_counts_df.loc[
                                  :, (positional_aa_counts_df != 0).any(axis=0)
                                  ]

        # start counting positions at 1
        positional_aa_counts_df.index = [idx + 1 for idx in list(positional_aa_counts_df.index)]

        return positional_aa_counts_df

    def _plot_gap_sizes(self, gap_size_df, x):
        file_path = self.result_path / f"{x}.html"

        gap_size_df["occurrence_total"] = gap_size_df.groupby("motif_size")["occurrence"].transform(sum)
        gap_size_df["occurrence_percentage"] = gap_size_df["occurrence"] / gap_size_df["occurrence_total"]

        x_label = x.replace("_", " ").capitalize()

        if self.motif_color_map is not None:
            color_discrete_map = self.motif_color_map
            color_discrete_sequence = None
        else:
            color_discrete_map = None
            color_discrete_sequence = self._get_color_discrete_sequence()

        gap_size_df = gap_size_df.sort_values(by=x)
        gap_size_df["motif_size"] = gap_size_df["motif_size"].astype(str)

        gap_size_fig = px.line(
            gap_size_df,
            x=x,
            y="occurrence_percentage",
            color="motif_size",
            color_discrete_map=color_discrete_map,
            color_discrete_sequence=color_discrete_sequence,
            category_orders=dict(motif_size=sorted([int(rate) for rate in gap_size_df["motif_size"].unique()])),
            template="plotly_white",
            markers=True,
            labels={
                x: x_label,
                "occurrence_percentage": "Percentage of motifs",
                "motif_size": "Motif size",
            },
        )
        gap_size_fig.layout.yaxis.tickformat = ',.0%'

        gap_size_fig.update_layout(font={"size": 14}, xaxis={"tickmode": "linear"})
        file_path = PlotlyUtil.write_image_to_file(gap_size_fig, file_path)

        return ReportOutput(
            path=file_path,
            name=f"Gap size between amino acids in high-precision high-recall motifs",
        )

    def _plot_positional_aa_counts(self, positional_aa_counts_df):
        file_path = self.result_path / f"positional_motif_frequencies.html"

        # reverse sort column names makes amino acids stack alphabetically in bar chart
        positional_aa_counts_df = positional_aa_counts_df[sorted(positional_aa_counts_df.columns)[::-1]]

        positional_aa_counts_fig = px.bar(
            positional_aa_counts_df,
            labels={
                "index": "Amino acid position",
                "value": "Frequency across motifs",
            },
            text="variable",
            color_discrete_map=PlotlyUtil.get_amino_acid_color_map(),
            template="plotly_white",
        )
        positional_aa_counts_fig.update_layout(
            showlegend=False, font={"size": 14}, xaxis={"tickmode": "linear"}
        )
        file_path = PlotlyUtil.write_image_to_file(positional_aa_counts_fig, file_path)
        return ReportOutput(
            path=file_path,
            name=f"Frequencies of amino acids found in the high-precision high-recall motifs",
        )

    def _get_color_discrete_sequence(self):
        return px.colors.qualitative.Pastel[:-1] + px.colors.qualitative.Set3

    def check_prerequisites(self):
        valid_encodings = [MotifEncoder.__name__]

        if self.dataset.encoded_data is None or self.dataset.encoded_data.info is None:
            logging.warning(
                "PositonalMotifFrequencies: the dataset is not encoded, skipping this report..."
            )
            return False
        elif self.dataset.encoded_data.encoding not in valid_encodings:
            logging.warning(
                f"PositonalMotifFrequencies: the dataset encoding ({self.dataset.encoded_data.encoding}) was not in the list of valid "
                f"encodings ({valid_encodings}), skipping this report..."
            )
            return False

        return True



import random
from pathlib import Path

from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.data_model.dataset.RepertoireDataset import RepertoireDataset
from immuneML.preprocessing.filters.Filter import Filter
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class SillyFilter(Filter):
    """
    This SillyFilter class is a placeholder for a real Preprocessor.
    It randomly selects a fraction of the repertoires to be removed from the dataset.


    **Specification arguments:**

    - fraction_to_keep (float): The fraction of repertoires to keep


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            preprocessing_sequences:
                my_preprocessing:
                    - step1:
                        SillyFilter:
                            fraction_to_remove: 0.8

    """

    def __init__(self, fraction_to_keep: float = None):
        super().__init__()
        self.fraction_to_keep = fraction_to_keep

    @classmethod
    def build_object(cls, **kwargs):
        # build_object is called early in the immuneML run, before the analysis takes place.
        # Its purpose is to fail early when a class is called incorrectly (checking parameters and dataset),
        # and provide user-friendly error messages.

        # ParameterValidator contains many utility functions for checking user parameters
        ParameterValidator.assert_type_and_value(kwargs['fraction_to_keep'], float, SillyFilter.__name__, 'fraction_to_keep', min_inclusive=0)

        return SillyFilter(**kwargs)

    def process_dataset(self, dataset: RepertoireDataset, result_path: Path, number_of_processes=1) -> RepertoireDataset:
        self.result_path = PathBuilder.build(result_path if result_path is not None else self.result_path)

        # utility function to ensure that the dataset type is RepertoireDataset
        self.check_dataset_type(dataset, [RepertoireDataset], SillyFilter.__name__)

        processed_dataset = self._create_random_dataset_subset(dataset)

        # utility function to ensure the remaining dataset is not empty
        self.check_dataset_not_empty(processed_dataset, SillyFilter.__name__)

        return processed_dataset

    def _create_random_dataset_subset(self, dataset):
        # Select some random fraction of identifiers, and use it to create a subset of the original dataset
        n_new_examples = round(dataset.get_example_count() * self.fraction_to_keep)
        new_example_indices = random.sample(range(dataset.get_example_count()), n_new_examples)

        preprocessed_dataset = dataset.make_subset(example_indices=new_example_indices,
                                                   path=self.result_path,
                                                   dataset_type=Dataset.SUBSAMPLED)

        return preprocessed_dataset

    def keeps_example_count(self):
        # Overwrite keeps_example_count to return False since some examples (repertoires) are removed
        return False




import abc
from typing import List

import pandas as pd
from sklearn.feature_extraction import DictVectorizer
from sklearn.preprocessing import StandardScaler

from immuneML.analysis.data_manipulation.NormalizationType import NormalizationType
from immuneML.caching.CacheHandler import CacheHandler
from immuneML.data_model.EncodedData import EncodedData
from immuneML.data_model.SequenceParams import RegionType
from immuneML.data_model.SequenceSet import ReceptorSequence
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.kmer_frequency.sequence_encoding.SequenceEncodingType import SequenceEncodingType
from immuneML.encodings.preprocessing.FeatureScaler import FeatureScaler
from immuneML.environment.SequenceType import SequenceType
from immuneML.util.EncoderHelper import EncoderHelper
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.ReadsType import ReadsType
from immuneML.util.ReflectionHandler import ReflectionHandler


class KmerFrequencyEncoder(DatasetEncoder):
    """
    The KmerFrequencyEncoder class encodes a repertoire, sequence or receptor by frequencies of k-mers it contains.
    A k-mer is a sequence of letters of length k into which an immune receptor sequence can be decomposed.
    K-mers can be defined in different ways, as determined by the sequence_encoding. If a dataset contains receptor
    sequences from multiple loci (e.g., TRA and TRB), the k-mer frequencies will be computed per locus and then combined
    into a single feature vector per sample. The k-mer frequencies can be normalized in different ways, as determined by
    the normalization_type. The design matrix can optionally be scaled to unit variance and/or to zero mean. The k-mer
    frequencies can be computed based on unique sequences (clonotypes) or taking into account the counts of the
    sequences in the repertoire.

    **Dataset type:**

    - SequenceDatasets

    - ReceptorDatasets

    - RepertoireDatasets


    **Specification arguments:**

    - sequence_encoding (:py:mod:`~immuneML.encodings.kmer_frequency.sequence_encoding.SequenceEncodingType`):
      Sequence encoding determines how the sequences are decomposed into k-mers. It includes:

      - CONTINUOUS_KMER: contiguous overlapping k-mers of length k (e.g., ACDE -> {ACD, CDE} for k=3) - default value
      - GAPPED_KMER: k-mers of length k_left + k_right with a gap of size between min_gap and max_gap in between
        (e.g., ACDE -> {AC, A.D, CD, C.E} for k_left=1, k_right=1, min_gap=0, max_gap=1)
      - IMGT_CONTINUOUS_KMER: contiguous k-mers of length k with IMGT positional information
        (e.g., AHCDE -> {'AHC_105', 'HCD_106', 'CDE_107'} for k=3)
      - IMGT_GAPPED_KMER: k-mers of length k_left + k_right with a gap of size between min_gap and max_gap in
        between, annotated by the starting IMGT position (e.g., AHCDE -> {'AH_105', 'HC_106', 'CD_107', 'DE_116',
        'A.C_105', 'H.D_106', 'C.E_107'} for k_left=1, k_right=1, min_gap=0, max_gap=1)
      - V_GENE_CONT_KMER: contiguous k-mers of length k, annotated by the V gene of the sequence they belong to
        (e.g., ACDE -> {V1-1_ACD, V1-1_CDE} for k=3 and V gene V1-1)
      - V_GENE_IMGT_KMER: contiguous k-mers of length k, annotated by the V gene of the sequence they belong to,
        annotated by the starting IMGT position (e.g., AHCDE -> {V1-1_AHC_105, V1-1_HCD_106, V1-1_CDE_107} for k=3 and
        V gene V1-1)
      - IDENTITY: the k-mers correspond to the original sequences

    - normalization_type (:py:mod:`~immuneML.analysis.data_manipulation.NormalizationType`): The way in which the
      k-mer frequencies should be normalized to unit norm; options are: binary, relative_frequency (also known as l1,
      default value), l2, max, none. For more information, see scikit-learn's documentation on
      `normalization <https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.normalize.html#sklearn.preprocessing.normalize>`_.

    - reads (:py:mod:`~immuneML.util.ReadsType`): Reads type signify whether the counts of the sequences in the
      repertoire will be taken into account. If :py:mod:`~immuneML.util.ReadsType.UNIQUE`, only unique sequences
      (clonotypes) are encoded, and if :py:mod:`~immuneML.util.ReadsType.ALL`, the sequence 'count' value is taken into
      account when determining the k-mer frequency. The default value for reads is unique.

    - k (int): Length of the k-mer (number of amino acids) when ungapped k-mers are used. The default value for k is 3.

    - k_left (int): When gapped k-mers are used, k_left indicates the length of the k-mer left of the gap. The default
      value for k_left is 1.

    - k_right (int): Same as k_left, but k_right determines the length of the k-mer right of the gap. The default value
      for k_right is 1.

    - min_gap (int): Minimum gap size when gapped k-mers are used. The default value for min_gap is 0.

    - max_gap: (int): Maximum gap size when gapped k-mers are used. The default value for max_gap is 0.

    - sequence_type (str): Whether to work with nucleotide or amino acid sequences. Amino acid sequences are the
      default. To work with either sequence type, the sequences of the desired type should be included in the datasets,
      e.g., listed under 'columns_to_load' parameter. By default, both types will be included if available. Valid values
      are: AMINO_ACID and NUCLEOTIDE.

    - scale_to_unit_variance (bool): whether to scale the design matrix after normalization to have unit variance per
      feature. Setting this argument to True might improve the subsequent classifier's performance depending on the type
      of the classifier. The default value for scale_to_unit_variance is true.

    - scale_to_zero_mean (bool): whether to scale the design matrix after normalization to have zero mean per feature.
      Setting this argument to True might improve the subsequent classifier's performance depending on the type of the
      classifier. However, if the original design matrix was sparse, setting this argument to True will destroy the
      sparsity and will increase the memory consumption. The default value for scale_to_zero_mean is false.

    - region_type (:py:mod:`~immuneML.data_model.SequenceParams.RegionType): the part of the receptor sequence to use
      in the analysis. The default value is IMGT_CDR3. Other values: IMGT_CDR1, IMGT_CDR2, IMGT_CDR3, IMGT_FR1,
      IMGT_FR2, IMGT_FR3, IMGT_FR4, IMGT_JUNCTION, FULL_SEQUENCE. Note that if an IMGT-based sequence encoding is
      used, the region_type has to be IMGT_CDR3 or IMGT_JUNCTION.


    **YAML specification:**

    .. indent with spaces
    .. code-block:: yaml

        definitions:
            encodings:
                my_continuous_kmer:
                    KmerFrequency:
                        normalization_type: RELATIVE_FREQUENCY
                        reads: UNIQUE
                        sequence_encoding: CONTINUOUS_KMER
                        sequence_type: NUCLEOTIDE
                        k: 3
                        scale_to_unit_variance: True
                        scale_to_zero_mean: True
                my_gapped_kmer:
                    KmerFrequency:
                        normalization_type: RELATIVE_FREQUENCY
                        reads: UNIQUE
                        sequence_encoding: GAPPED_KMER
                        sequence_type: AMINO_ACID
                        k_left: 2
                        k_right: 2
                        min_gap: 1
                        max_gap: 3
                        scale_to_unit_variance: True
                        scale_to_zero_mean: False

    """

    STEP_ENCODED = "encoded"
    STEP_VECTORIZED = "vectorized"
    STEP_NORMALIZED = "normalized"
    STEP_SCALED = "scaled"

    dataset_mapping = {
        "RepertoireDataset": "KmerFreqRepertoireEncoder",
        "SequenceDataset": "KmerFreqSequenceEncoder",
        "ReceptorDataset": "KmerFreqReceptorEncoder"
    }

    def __init__(self, normalization_type: NormalizationType, reads: ReadsType, sequence_encoding: SequenceEncodingType,
                 k: int = 0, k_left: int = 0, k_right: int = 0, min_gap: int = 0, max_gap: int = 0,
                 name: str = None, scale_to_unit_variance: bool = False,
                 scale_to_zero_mean: bool = False, sequence_type: SequenceType = None, include_v_genes: str = None,
                 region_type: RegionType = RegionType.IMGT_CDR3):
        super().__init__(name=name)
        self.normalization_type = normalization_type
        self.reads = reads
        self.sequence_encoding = sequence_encoding
        self.include_v_genes = include_v_genes
        self.sequence_type = sequence_type
        self.region_type = region_type
        self.k = k
        self.k_left = k_left
        self.k_right = k_right
        self.min_gap = min_gap
        self.max_gap = max_gap
        self.scale_to_unit_variance = scale_to_unit_variance
        self.scale_to_zero_mean = scale_to_zero_mean
        self.scaler = None
        self.vectorizer = None

    @staticmethod
    def _prepare_parameters(normalization_type: str, reads: str, sequence_encoding: str, k: int = 0, k_left: int = 0,
                            k_right: int = 0, min_gap: int = 0, max_gap: int = 0, name: str = None,
                            scale_to_unit_variance: bool = False, scale_to_zero_mean: bool = False,
                            sequence_type: str = SequenceType.AMINO_ACID.name,
                            region_type: str = RegionType.IMGT_CDR3.name):

        location = KmerFrequencyEncoder.__name__

        ParameterValidator.assert_in_valid_list(normalization_type.upper(), [item.name for item in NormalizationType], location, "normalization_type")
        ParameterValidator.assert_in_valid_list(reads.upper(), [item.name for item in ReadsType], location, "reads")
        ParameterValidator.assert_in_valid_list(sequence_encoding.upper(), [item.name for item in SequenceEncodingType], location,
                                                "sequence_encoding")
        ParameterValidator.assert_type_and_value(scale_to_zero_mean, bool, location, "scale_to_zero_mean")
        ParameterValidator.assert_type_and_value(scale_to_unit_variance, bool, location, "scale_to_unit_variance")
        ParameterValidator.assert_type_and_value(sequence_type, str, location, 'sequence_type')
        ParameterValidator.assert_in_valid_list(sequence_type.upper(), [st.name for st in SequenceType], location, 'sequence_type')

        if "IMGT" in sequence_encoding.upper():
            assert sequence_type.upper() == SequenceType.AMINO_ACID.name, f"{location}: for IMGT-based k-mer frequency encoding (here: " \
                                                                          f"{sequence_encoding.upper()}), sequence type has to be 'amino_acid'."

        vars_to_check = {"k": k, "k_left": k_left, "k_right": k_right, "min_gap": min_gap, "max_gap": max_gap}
        for param in vars_to_check.keys():
            ParameterValidator.assert_type_and_value(vars_to_check[param], int, location, param, min_inclusive=0)

        if "gap" in sequence_encoding.lower():
            assert k_left != 0 and k_right != 0, f"KmerFrequencyEncoder: sequence encoding {sequence_encoding} was chosen, but k_left " \
                                                 f"({k_left}) or k_right ({k_right}) have to be set and larger than 0."

        return {
            "normalization_type": NormalizationType[normalization_type.upper()],
            "reads": ReadsType[reads.upper()],
            "sequence_encoding": SequenceEncodingType[sequence_encoding.upper()],
            "name": name,
            "scale_to_zero_mean": scale_to_zero_mean, "scale_to_unit_variance": scale_to_unit_variance,
            'sequence_type': SequenceType[sequence_type.upper()],
            'region_type': RegionType[region_type.upper()],
            **vars_to_check
        }

    @staticmethod
    def build_object(dataset=None, **params):
        EncoderHelper.check_dataset_type_available_in_mapping(dataset, KmerFrequencyEncoder)

        prepared_params = KmerFrequencyEncoder._prepare_parameters(**params)
        encoder = ReflectionHandler.get_class_by_name(KmerFrequencyEncoder.dataset_mapping[dataset.__class__.__name__],
                                                      "kmer_frequency/")(**prepared_params)

        return encoder

    def encode(self, dataset, params: EncoderParams):

        cache_params = self._prepare_caching_params(dataset, params)

        encoded_dataset = CacheHandler.memo_by_params(cache_params, lambda: self._encode_new_dataset(dataset, params))

        EncoderHelper.sync_encoder_with_cache(cache_params, lambda: {'vectorizer': self.vectorizer, 'scaler': self.scaler}, self,
                                              ['vectorizer', 'scaler'])

        return encoded_dataset

    def _prepare_caching_params(self, dataset, params: EncoderParams, step: str = ""):
        return (("dataset_identifier", dataset.identifier),
                ("labels", tuple(params.label_config.get_labels_by_name())),
                ("encoding", KmerFrequencyEncoder.__name__),
                ("learn_model", params.learn_model),
                ("step", step),
                ("encoding_params", tuple(vars(self).items())))

    def _encode_data(self, dataset, params: EncoderParams) -> EncodedData:
        encoded_example_list, example_ids, encoded_labels = CacheHandler.memo_by_params(
            self._prepare_caching_params(dataset, params, KmerFrequencyEncoder.STEP_ENCODED),
            lambda: self._encode_examples(dataset, params))

        self._initialize_vectorizer(params)
        vectorized_examples = self._vectorize_encoded(examples=encoded_example_list, params=params,
                                                      vectorizer=self.vectorizer)
        feature_names = self.vectorizer.feature_names_
        normalized_examples = FeatureScaler.normalize(vectorized_examples, self.normalization_type)

        if self.scale_to_unit_variance:
            examples = self.scale_normalized(params, dataset, normalized_examples)
        else:
            examples = normalized_examples

        encoded_data = EncodedData(examples=examples,
                                   labels=encoded_labels,
                                   feature_names=feature_names,
                                   example_ids=example_ids,
                                   feature_annotations=pd.DataFrame({'feature': feature_names}),
                                   encoding=KmerFrequencyEncoder.__name__,
                                   info={"sequence_type": self.sequence_type, 'region_type': self.region_type})

        return encoded_data

    def scale_normalized(self, params, dataset, normalized_examples):

        if params.learn_model:
            self.scaler = StandardScaler(with_mean=self.scale_to_zero_mean)
            examples = CacheHandler.memo_by_params(
                self._prepare_caching_params(dataset, params, step=KmerFrequencyEncoder.STEP_SCALED),
                lambda: FeatureScaler.standard_scale_fit(self.scaler, normalized_examples, with_mean=self.scale_to_zero_mean))
        else:
            examples = CacheHandler.memo_by_params(
                self._prepare_caching_params(dataset, params, step=KmerFrequencyEncoder.STEP_SCALED),
                lambda: FeatureScaler.standard_scale(self.scaler, normalized_examples, with_mean=self.scale_to_zero_mean))

        return examples

    @abc.abstractmethod
    def _encode_new_dataset(self, dataset, params: EncoderParams):
        pass

    @abc.abstractmethod
    def _encode_examples(self, dataset, params: EncoderParams):
        pass

    def _initialize_vectorizer(self, params: EncoderParams):
        if self.vectorizer is None or params.learn_model:
            self.vectorizer = DictVectorizer(sparse=True, dtype=float)

    def _vectorize_encoded(self, examples: list, params: EncoderParams, vectorizer: DictVectorizer):

        if params.learn_model:
            vectorized_examples = vectorizer.fit_transform(examples)
        else:
            vectorized_examples = vectorizer.transform(examples)

        return vectorized_examples

    def _prepare_sequence_encoder(self):
        class_name = self.sequence_encoding.value
        sequence_encoder = ReflectionHandler.get_class_by_name(class_name, "encodings/")
        return sequence_encoder

    def _encode_sequence(self, sequence: ReceptorSequence, params: EncoderParams, sequence_encoder, counts, encode_locus: bool):
        params.model = vars(self)
        features = sequence_encoder.encode_sequence(sequence, params, encode_locus)
        if features is not None:
            for i in features:
                if self.reads == ReadsType.UNIQUE:
                    counts[i] += 1
                elif self.reads == ReadsType.ALL:
                    counts[i] += sequence.duplicate_count
        return counts

    @abc.abstractmethod
    def _encode_locus(self, dataset):
        pass

    def get_additional_files(self) -> List[str]:
        return []

